from collections import defaultdict
from gettext import gettext as _
import os
import logging
import shutil
import tempfile

import createrepo_c as cr
import libcomps

from django.conf import settings
from django.core.files import File
from django.db.models import Q

from pulpcore.plugin.models import (
    AsciiArmoredDetachedSigningService,
    ContentArtifact,
    RepositoryVersion,
    ProgressReport,
    PublishedArtifact,
    PublishedMetadata,
    RepositoryContent,
)

from pulp_rpm.app.comps import dict_to_strdict
from pulp_rpm.app.constants import ALLOWED_CHECKSUM_ERROR_MSG, CHECKSUM_TYPES, PACKAGES_DIRECTORY
from pulp_rpm.app.kickstart.treeinfo import PulpTreeInfo, TreeinfoData
from pulp_rpm.app.models import (
    DistributionTree,
    Modulemd,
    ModulemdDefaults,
    ModulemdObsolete,
    Package,
    PackageCategory,
    PackageEnvironment,
    PackageGroup,
    PackageLangpacks,
    RepoMetadataFile,
    RpmPublication,
    UpdateRecord,
)

log = logging.getLogger(__name__)

REPODATA_PATH = "repodata"


class PublicationData:
    """
    Encapsulates data relative to publication.

    Attributes:
        publication (pulpcore.plugin.models.Publication): A Publication to populate.
        sub_repos (list): A list of tuples with sub_repos data.
        repomdrecords (list): A list of tuples with repomdrecords data.

    """

    def __init__(self, publication):
        """
        Setting Publication data.

        Args:
            publication (pulpcore.plugin.models.Publication): A Publication to populate.

        """
        self.publication = publication
        self.sub_repos = []
        self.repomdrecords = []

    def prepare_metadata_files(self, content, folder=None):
        """
        Copies metadata files from the Artifact storage.

        Args:
            content (pulpcore.plugin.models.Content): content set.

        Keyword Args:
            folder(str): name of the directory.

        Returns:
            repomdrecords (list): A list of tuples with repomdrecords data.

        """
        repomdrecords = []
        repo_metadata_files = RepoMetadataFile.objects.filter(pk__in=content).prefetch_related(
            "contentartifact_set"
        )

        for repo_metadata_file in repo_metadata_files:
            if repo_metadata_file.unsupported_metadata_type:
                # Normally these types are not synced in the first place, we skip them here, since
                # they might still exist in old repo versions from before we started excluding them.
                continue
            content_artifact = repo_metadata_file.contentartifact_set.get()
            current_file = content_artifact.artifact.file.file
            path = content_artifact.relative_path.split("/")[-1]
            if repo_metadata_file.checksum in path:
                # filenames can be checksum-xxxx.yyy - but can also be checksum-mmm-nnn-ooo.yyy
                # Some old repos might have filename of checksum only so we need to take care
                # of that too
                # split off the checksum, keep the rest
                filename = path.split("-")[1:]
                if filename:
                    path = "-".join(filename)
            if folder:
                path = os.path.join(folder, path)
            with open(path, "wb") as new_file:
                shutil.copyfileobj(current_file, new_file)
                repomdrecords.append((repo_metadata_file.data_type, new_file.name, None))

        return repomdrecords

    def publish_artifacts(self, content, prefix=""):
        """
        Publish artifacts.

        Args:
            content (pulpcore.plugin.models.Content): content set.
            prefix (str): a relative path prefix for the published artifact

        """
        published_artifacts = []

        # Special case for Packages
        contentartifact_qs = ContentArtifact.objects.filter(content__in=content).filter(
            content__pulp_type=Package.get_pulp_type()
        )

        rel_path_mapping = defaultdict(list)
        # Some Suboptimal Repos have the 'same' artifact living in multiple places.
        # Specifically, the same NEVRA, in more than once place, **with different checksums**
        # (since if all that was different was location_href there would be only one
        # ContentArtifact in the first place).
        #
        # pulp_rpm wants to publish a 'canonical' repository-layout, under which an RPM
        # "name-version-release-arch" appears at "Packages/n/name-version-release-arch.rpm".
        # Because the assumption is that Packages don't "own" their path, only the filename
        # is kept as relative_path.
        #
        # In this case, we have to pick one - which is essentially what the rest of the RPM
        # Ecosystem does when faced with the impossible. This code takes the one with the
        # most recent build time which is the same heuristic used by Yum/DNF/Zypper.
        #
        # Note that this only impacts user-created publications, which produce the "standard"
        # RPM layout of repo/Packages/f/foo.rpm. A publication created by mirror-sync retains
        # whatever layout their "upstream" repo-metadata dictates.
        fields = ["pk", "relative_path", "content__rpm_package__time_build"]
        for content_artifact in contentartifact_qs.values(*fields).iterator():
            relative_path = content_artifact["relative_path"]
            time_build = content_artifact["content__rpm_package__time_build"]

            relative_path = os.path.join(
                prefix, PACKAGES_DIRECTORY, relative_path.lower()[0], relative_path
            )
            rel_path_mapping[relative_path].append((content_artifact["pk"], time_build))

        for rel_path, content_artifacts in rel_path_mapping.items():
            # sort the content artifacts by when the package was built
            if len(content_artifacts) > 1:
                content_artifacts.sort(key=lambda p: p[1], reverse=True)
                log.warning(
                    "Duplicate packages found competing for {path}, selected the one with "
                    "the most recent build time, excluding {others} others.".format(
                        path=rel_path, others=len(content_artifacts[1:])
                    )
                )

            # Only add the first one (the one with the highest build time)
            published_artifacts.append(
                PublishedArtifact(
                    relative_path=rel_path,
                    publication=self.publication,
                    content_artifact_id=content_artifacts[0][0],
                )
            )

        # Handle everything else
        is_treeinfo = Q(relative_path__in=["treeinfo", ".treeinfo"])
        unpublishable_types = Q(
            content__pulp_type__in=[
                RepoMetadataFile.get_pulp_type(),
                Modulemd.get_pulp_type(),
                ModulemdDefaults.get_pulp_type(),
                # already dealt with
                Package.get_pulp_type(),
            ]
        )

        contentartifact_qs = (
            ContentArtifact.objects.filter(content__in=content)
            .exclude(unpublishable_types)
            .exclude(is_treeinfo)
        )

        for content_artifact in contentartifact_qs.values("pk", "relative_path").iterator():
            published_artifacts.append(
                PublishedArtifact(
                    relative_path=content_artifact["relative_path"],
                    publication=self.publication,
                    content_artifact_id=content_artifact["pk"],
                )
            )

        PublishedArtifact.objects.bulk_create(published_artifacts, batch_size=2000)

    def handle_sub_repos(self, distribution_tree):
        """
        Get sub-repo content and publish them.

        Args:
            distribution_tree (pulp_rpm.models.DistributionTree): A distribution_tree object.

        """
        original_treeinfo_content_artifact = distribution_tree.contentartifact_set.get(
            relative_path__in=[".treeinfo", "treeinfo"]
        )
        orig_artifact = original_treeinfo_content_artifact.artifact
        artifact_file = orig_artifact.pulp_domain.get_storage().open(orig_artifact.file.name)
        with tempfile.NamedTemporaryFile("wb", dir=".") as temp_file:
            shutil.copyfileobj(artifact_file, temp_file)
            temp_file.flush()
            treeinfo = PulpTreeInfo()
            treeinfo.load(f=temp_file.name)
            treeinfo_data = TreeinfoData(treeinfo.parsed_sections())

            # rewrite the treeinfo file such that the variant repository and package location
            # is a relative subtree
            treeinfo.rewrite_subrepo_paths(treeinfo_data)

            # TODO: better way to do this?
            main_variant = treeinfo.original_parser._sections.get("general", {}).get(
                "variant", None
            )
            treeinfo_file = tempfile.NamedTemporaryFile(dir=".")
            treeinfo.dump(treeinfo_file.name, main_variant=main_variant)
            with open(treeinfo_file.name, "rb") as treeinfo_fd:
                PublishedMetadata.create_from_file(
                    relative_path=original_treeinfo_content_artifact.relative_path,
                    publication=self.publication,
                    file=File(treeinfo_fd),
                )
        artifact_file.close()
        relations = ["addon", "variant"]
        for relation in relations:
            addons_or_variants = getattr(distribution_tree, f"{relation}s").all()
            for addon_or_variant in addons_or_variants:
                if not addon_or_variant.repository:
                    # a variant of the main repo
                    continue
                repository = addon_or_variant.repository.cast()
                repository_version = repository.latest_version()

                if repository_version and repository.user_hidden:
                    addon_or_variant_id = getattr(addon_or_variant, f"{relation}_id")
                    self.sub_repos.append(
                        (
                            addon_or_variant_id,
                            repository_version.content,
                            repository.original_checksum_types,
                        )
                    )

    def populate(self):
        """
        Populate a publication.

        Create published artifacts for a publication.

        """
        main_content = self.publication.repository_version.content
        self.repomdrecords = self.prepare_metadata_files(main_content)

        self.publish_artifacts(main_content)

        distribution_trees = DistributionTree.objects.filter(pk__in=main_content).prefetch_related(
            "addons",
            "variants",
            "addons__repository",
            "variants__repository",
            "contentartifact_set",
        )

        for distribution_tree in distribution_trees:
            self.handle_sub_repos(distribution_tree)

        for name, content, checksum_types in self.sub_repos:
            os.mkdir(name)
            setattr(self, f"{name}_content", content)
            setattr(self, f"{name}_checksums", checksum_types)
            setattr(self, f"{name}_repomdrecords", self.prepare_metadata_files(content, name))
            self.publish_artifacts(content, prefix=name)


def get_checksum_type(name, checksum_types, default=CHECKSUM_TYPES.SHA256):
    """
    Get checksum algorithm for publishing metadata.

    Args:
        name (str): Name of the metadata type.
        checksum_types (dict): Checksum types for metadata and packages.
    Kwargs:
        default: The checksum type used if there is no specified nor original checksum type.
    """
    original = checksum_types.get("original")
    metadata = checksum_types.get("metadata")
    checksum_type = metadata if metadata else original.get(name, default)
    # "sha" -> "SHA" -> "CHECKSUM_TYPES.SHA" -> "sha1"
    normalized_checksum_type = getattr(CHECKSUM_TYPES, checksum_type.upper())
    return normalized_checksum_type


def cr_checksum_type_from_string(checksum_type):
    """
    Convert checksum type from string to createrepo_c enum variant.
    """
    # "sha1" -> "SHA1" -> "cr.SHA1"
    return getattr(cr, checksum_type.upper())


def publish(
    repository_version_pk,
    metadata_signing_service=None,
    checksum_types=None,
    sqlite_metadata=False,
    repo_config=None,
):
    """
    Create a Publication based on a RepositoryVersion.

    Args:
        repository_version_pk (str): Create a publication from this repository version.
        metadata_signing_service (pulpcore.app.models.AsciiArmoredDetachedSigningService):
            A reference to an associated signing service.
        checksum_types (dict): Checksum types for metadata and packages.
        sqlite_metadata (bool): Whether to generate metadata files in sqlite format.
        repo_config (JSON): repo config that will be served by distribution

    """
    repository_version = RepositoryVersion.objects.get(pk=repository_version_pk)
    repository = repository_version.repository.cast()
    checksum_types = checksum_types or {}

    if metadata_signing_service:
        metadata_signing_service = AsciiArmoredDetachedSigningService.objects.get(
            pk=metadata_signing_service
        )

    checksum_types["original"] = repository.original_checksum_types

    log.info(
        _("Publishing: repository={repo}, version={version}").format(
            repo=repository.name,
            version=repository_version.number,
        )
    )
    with tempfile.TemporaryDirectory(dir="."):
        with RpmPublication.create(repository_version) as publication:
            publication.metadata_checksum_type = get_checksum_type("primary", checksum_types)
            publication.package_checksum_type = (
                checksum_types.get("package") or publication.metadata_checksum_type
            )

            if sqlite_metadata:
                publication.sqlite_metadata = True

            publication.repo_config = repo_config

            publication_data = PublicationData(publication)
            publication_data.populate()

            total_repos = 1 + len(publication_data.sub_repos)
            pb_data = dict(
                message="Generating repository metadata",
                code="publish.generating_metadata",
                total=total_repos,
            )
            with ProgressReport(**pb_data) as publish_pb:
                content = publication.repository_version.content

                # Main repo
                generate_repo_metadata(
                    content,
                    publication,
                    checksum_types,
                    publication_data.repomdrecords,
                    metadata_signing_service=metadata_signing_service,
                )
                publish_pb.increment()

                for sub_repo in publication_data.sub_repos:
                    name = sub_repo[0]
                    checksum_types["original"] = getattr(publication_data, f"{name}_checksums")
                    content = getattr(publication_data, f"{name}_content")
                    extra_repomdrecords = getattr(publication_data, f"{name}_repomdrecords")
                    generate_repo_metadata(
                        content,
                        publication,
                        checksum_types,
                        extra_repomdrecords,
                        name,
                        metadata_signing_service=metadata_signing_service,
                    )
                    publish_pb.increment()

            log.info(_("Publication: {publication} created").format(publication=publication.pk))

            return publication


def generate_repo_metadata(
    content,
    publication,
    checksum_types,
    extra_repomdrecords,
    sub_folder=None,
    metadata_signing_service=None,
):
    """
    Creates a repomd.xml file.

    Args:
        content(app.models.Content): content set
        publication(pulpcore.plugin.models.Publication): the publication
        extra_repomdrecords(list): list with data relative to repo metadata files
        sub_folder(str): name of the folder for sub repos
        metadata_signing_service (pulpcore.app.models.AsciiArmoredDetachedSigningService):
            A reference to an associated signing service.

    """
    cwd = os.getcwd()
    repodata_path = REPODATA_PATH
    has_modules = False
    has_comps = False
    package_checksum_type = checksum_types.get("package")

    if sub_folder:
        cwd = os.path.join(cwd, sub_folder)
        repodata_path = os.path.join(sub_folder, repodata_path)

    if package_checksum_type and package_checksum_type not in settings.ALLOWED_CONTENT_CHECKSUMS:
        raise ValueError(
            "Repository contains disallowed package checksum type '{}', "
            "thus can't be published. {}".format(package_checksum_type, ALLOWED_CHECKSUM_ERROR_MSG)
        )

    # Prepare metadata files
    repomd_path = os.path.join(cwd, "repomd.xml")
    pri_xml_path = os.path.join(cwd, "primary.xml.gz")
    fil_xml_path = os.path.join(cwd, "filelists.xml.gz")
    oth_xml_path = os.path.join(cwd, "other.xml.gz")
    upd_xml_path = os.path.join(cwd, "updateinfo.xml.gz")
    mod_yml_path = os.path.join(cwd, "modules.yaml")
    comps_xml_path = os.path.join(cwd, "comps.xml")

    pri_xml = cr.PrimaryXmlFile(pri_xml_path, compressiontype=cr.GZ)
    fil_xml = cr.FilelistsXmlFile(fil_xml_path, compressiontype=cr.GZ)
    oth_xml = cr.OtherXmlFile(oth_xml_path, compressiontype=cr.GZ)
    upd_xml = None

    if publication.sqlite_metadata:
        pri_db_path = os.path.join(cwd, "primary.sqlite")
        fil_db_path = os.path.join(cwd, "filelists.sqlite")
        oth_db_path = os.path.join(cwd, "other.sqlite")
        pri_db = cr.PrimarySqlite(pri_db_path)
        fil_db = cr.FilelistsSqlite(fil_db_path)
        oth_db = cr.OtherSqlite(oth_db_path)

    # We want to support publishing with a different checksum type than the one built-in to the
    # package itself, so we need to get the correct checksums somehow if there is an override.
    # We must also take into consideration that if the package has not been downloaded the only
    # checksum that is available is the one built-in.
    #
    # Since this lookup goes from Package->Content->ContentArtifact->Artifact, performance is a
    # challenge. We use ContentArtifact as our starting point because it enables us to work with
    # simple foreign keys and avoid messing with the many-to-many relationship, which doesn't
    # work with select_related() and performs poorly with prefetch_related(). This is fine
    # because we know that Packages should only ever have one artifact per content.
    fields = [
        "content_id",
        "content__rpm_package__checksum_type",
        "content__rpm_package__pkgId",
    ]
    artifact_checksum = None
    if package_checksum_type:
        package_checksum_type = package_checksum_type.lower()
        artifact_checksum = f"artifact__{package_checksum_type}"
        fields.append(artifact_checksum)

    contentartifact_qs = ContentArtifact.objects.filter(
        content__in=content, content__pulp_type=Package.get_pulp_type()
    ).values(*fields)

    pkg_to_hash = {}
    for ca in contentartifact_qs.iterator():
        if package_checksum_type:
            pkgid = ca.get(artifact_checksum, None)

        if not package_checksum_type or not pkgid:
            if ca["content__rpm_package__checksum_type"] not in settings.ALLOWED_CONTENT_CHECKSUMS:
                raise ValueError(
                    "Package with pkgId {} as content unit {} contains forbidden checksum type "
                    "'{}', thus can't be published. {}".format(
                        ca["content__rpm_package__pkgId"],
                        ca["content_id"],
                        ca["content__rpm_package__checksum_type"],
                        ALLOWED_CHECKSUM_ERROR_MSG,
                    )
                )
            package_checksum_type = ca["content__rpm_package__checksum_type"]
            pkgid = ca["content__rpm_package__pkgId"]

        pkg_to_hash[ca["content_id"]] = (package_checksum_type, pkgid)

    # TODO: this is meant to be a !! *temporary* !! fix for
    # https://github.com/pulp/pulp_rpm/issues/2407
    pkg_pks_to_ignore = set()
    latest_build_time_by_nevra = defaultdict(list)
    packages = Package.objects.filter(pk__in=content)
    for pkg in packages.only(
        "pk", "name", "epoch", "version", "release", "arch", "time_build"
    ).iterator():
        latest_build_time_by_nevra[pkg.nevra].append((pkg.time_build, pkg.pk))
    for nevra, pkg_data in latest_build_time_by_nevra.items():
        # sort the packages by when they were built
        if len(pkg_data) > 1:
            pkg_data.sort(key=lambda p: p[0], reverse=True)
            pkg_pks_to_ignore |= set(entry[1] for entry in pkg_data[1:])
            log.warning(
                "Duplicate packages found competing for NEVRA {nevra}, selected the one with "
                "the most recent build time, excluding {others} others.".format(
                    nevra=nevra, others=len(pkg_data[1:])
                )
            )

    total_packages = packages.count() - len(pkg_pks_to_ignore)

    pri_xml.set_num_of_pkgs(total_packages)
    fil_xml.set_num_of_pkgs(total_packages)
    oth_xml.set_num_of_pkgs(total_packages)

    if settings.RPM_METADATA_USE_REPO_PACKAGE_TIME:
        # gather the times the packages were added to the repo
        repo_content = (
            RepositoryContent.objects.filter(
                repository=publication.repository,
                version_added__number__lte=publication.repository_version.number,
            )
            .exclude(version_removed__number__lte=publication.repository_version.number)
            .values_list("content", "pulp_created")
        )
        repo_pkg_times = {pk: created.timestamp() for pk, created in repo_content}

    # Process all packages
    for package in packages.order_by("name", "evr").iterator():
        if package.pk in pkg_pks_to_ignore:  # Temporary!
            continue
        pkg = package.to_createrepo_c()

        # rewrite the checksum and checksum type with the desired ones
        (checksum, pkgId) = pkg_to_hash[package.pk]
        pkg.checksum_type = checksum
        pkg.pkgId = pkgId

        pkg_filename = os.path.basename(package.location_href)
        # this can cause an issue when two same RPM package names appears
        # a/name1.rpm b/name1.rpm
        pkg.location_href = os.path.join(PACKAGES_DIRECTORY, pkg_filename[0].lower(), pkg_filename)

        if settings.RPM_METADATA_USE_REPO_PACKAGE_TIME:
            pkg.time_file = repo_pkg_times[package.pk]

        pri_xml.add_pkg(pkg)
        fil_xml.add_pkg(pkg)
        oth_xml.add_pkg(pkg)
        if publication.sqlite_metadata:
            pri_db.add_pkg(pkg)
            fil_db.add_pkg(pkg)
            oth_db.add_pkg(pkg)

    # Process update records
    update_records = UpdateRecord.objects.filter(pk__in=content).order_by("id", "digest")
    for update_record in update_records.iterator():
        if not upd_xml:
            upd_xml = cr.UpdateInfoXmlFile(upd_xml_path, compressiontype=cr.GZ)
        upd_xml.add_chunk(cr.xml_dump_updaterecord(update_record.to_createrepo_c()))

    # Process modulemd, modulemd_defaults and obsoletes
    with open(mod_yml_path, "ab") as mod_yml:
        modulemds = Modulemd.objects.filter(pk__in=content).order_by(*Modulemd.natural_key_fields())
        for modulemd in modulemds.iterator():
            mod_yml.write(modulemd.snippet.encode())
            mod_yml.write(b"\n")
            has_modules = True
        modulemd_defaults = ModulemdDefaults.objects.filter(pk__in=content).order_by(
            *ModulemdDefaults.natural_key_fields()
        )
        for default in modulemd_defaults.iterator():
            mod_yml.write(default.snippet.encode())
            mod_yml.write(b"\n")
            has_modules = True
        modulemd_obsoletes = ModulemdObsolete.objects.filter(pk__in=content).order_by(
            *ModulemdObsolete.natural_key_fields()
        )
        for obsolete in modulemd_obsoletes.iterator():
            mod_yml.write(obsolete.snippet.encode())
            mod_yml.write(b"\n")
            has_modules = True

    # Process comps
    comps = libcomps.Comps()
    for pkg_grp in PackageGroup.objects.filter(pk__in=content).order_by("id").iterator():
        group = pkg_grp.pkg_grp_to_libcomps()
        comps.groups.append(group)
        has_comps = True
    for pkg_cat in PackageCategory.objects.filter(pk__in=content).order_by("id").iterator():
        cat = pkg_cat.pkg_cat_to_libcomps()
        comps.categories.append(cat)
        has_comps = True
    for pkg_env in PackageEnvironment.objects.filter(pk__in=content).order_by("id").iterator():
        env = pkg_env.pkg_env_to_libcomps()
        comps.environments.append(env)
        has_comps = True
    package_langpacks = PackageLangpacks.objects.filter(pk__in=content).order_by(
        *PackageLangpacks.natural_key_fields()
    )
    for pkg_lng in package_langpacks.iterator():
        comps.langpacks = dict_to_strdict(pkg_lng.matches)
        has_comps = True

    comps.toxml_f(
        comps_xml_path,
        xml_options={
            "default_explicit": True,
            "empty_groups": True,
            "empty_packages": True,
            "uservisible_explicit": True,
        },
    )

    pri_xml.close()
    fil_xml.close()
    oth_xml.close()
    if upd_xml:
        upd_xml.close()

    repomd = cr.Repomd()
    # If the repository is empty, use a revision of 0
    # See: https://pulp.plan.io/issues/9402
    if not content.exists():
        repomd.revision = "0"

    if publication.sqlite_metadata:
        repomdrecords = [
            ("primary", pri_xml_path, pri_db),
            ("filelists", fil_xml_path, fil_db),
            ("other", oth_xml_path, oth_db),
            ("primary_db", pri_db_path, None),
            ("filelists_db", fil_db_path, None),
            ("other_db", oth_db_path, None),
        ]
    else:
        repomdrecords = [
            ("primary", pri_xml_path, None),
            ("filelists", fil_xml_path, None),
            ("other", oth_xml_path, None),
        ]

    if upd_xml:
        repomdrecords.append(("updateinfo", upd_xml_path, None))

    if has_modules:
        repomdrecords.append(("modules", mod_yml_path, None))

    if has_comps:
        repomdrecords.append(("group", comps_xml_path, None))

    repomdrecords.extend(extra_repomdrecords)

    sqlite_files = ("primary_db", "filelists_db", "other_db")

    for name, path, db_to_update in repomdrecords:
        record = cr.RepomdRecord(name, path)
        checksum_type = cr_checksum_type_from_string(
            get_checksum_type(name, checksum_types, default=publication.metadata_checksum_type)
        )
        if name in sqlite_files:
            record_bz = record.compress_and_fill(checksum_type, cr.BZ2)
            record_bz.type = name
            record_bz.rename_file()
            path = record_bz.location_href.split("/")[-1]
            repomd.set_record(record_bz)
        else:
            record.fill(checksum_type)
            if db_to_update:
                db_to_update.dbinfo_update(record.checksum)
                db_to_update.close()
            record.rename_file()
            path = record.location_href.split("/")[-1]
            repomd.set_record(record)

        if sub_folder:
            path = os.path.join(sub_folder, path)

        with open(path, "rb") as repodata_fd:
            PublishedMetadata.create_from_file(
                relative_path=os.path.join(repodata_path, os.path.basename(path)),
                publication=publication,
                file=File(repodata_fd),
            )

    with open(repomd_path, "w") as repomd_f:
        repomd_f.write(repomd.xml_dump())

    if metadata_signing_service:
        signing_service = AsciiArmoredDetachedSigningService.objects.get(
            pk=metadata_signing_service
        )
        sign_results = signing_service.sign(repomd_path)

        # publish a signed file
        with open(sign_results["file"], "rb") as signed_file_fd:
            PublishedMetadata.create_from_file(
                relative_path=os.path.join(repodata_path, os.path.basename(sign_results["file"])),
                publication=publication,
                file=File(signed_file_fd),
            )

        # publish a detached signature
        with open(sign_results["signature"], "rb") as signature_fd:
            PublishedMetadata.create_from_file(
                relative_path=os.path.join(
                    repodata_path, os.path.basename(sign_results["signature"])
                ),
                publication=publication,
                file=File(signature_fd),
            )

        # publish a public key required for further verification
        pubkey_name = "repomd.xml.key"
        with open(pubkey_name, "wb+") as f:
            f.write(signing_service.public_key.encode("utf-8"))
            f.flush()
            # important! as the file has already been opened and used, it will be treated as a
            # cursor and when calculating the checksum it will calculate the checksum of nothing.
            f.seek(0)
            PublishedMetadata.create_from_file(
                relative_path=os.path.join(repodata_path, pubkey_name),
                publication=publication,
                file=File(f),
            )
    else:
        with open(repomd_path, "rb") as repomd_fd:
            PublishedMetadata.create_from_file(
                relative_path=os.path.join(repodata_path, os.path.basename(repomd_path)),
                publication=publication,
                file=File(repomd_fd),
            )



"""Verify whether package manager, yum/dnf, can consume content from Pulp."""
import pytest
import subprocess
import itertools

from pulp_rpm.tests.functional.constants import (
    # REPO_WITH_XML_BASE_URL,
    RPM_UNSIGNED_FIXTURE_URL,
)


dnf_installed = subprocess.run(("which", "dnf")).returncode == 0


@pytest.fixture
def dnf_config_add_repo():
    added_repos = []

    def _add_repo(distribution, has_signing_service=False):
        subprocess.run(("sudo", "dnf", "config-manager", "--add-repo", distribution.base_url))
        repo_id = "*{}_".format(distribution.base_path)
        args = ["sudo", "dnf", "config-manager", "--save", f"--setopt={repo_id}.gpgcheck=0"]
        if has_signing_service:
            public_key_url = f"{distribution.base_url}repodata/repomd.xml.key"
            args.extend(
                (
                    f"--setopt={repo_id}.repo_gpgcheck=1",
                    f"--setopt={repo_id}.gpgkey={public_key_url}",
                )
            )

        subprocess.run(args + [repo_id])
        added_repos.append(repo_id)

    yield _add_repo
    for repo_id in added_repos:
        subprocess.run(("sudo", "dnf", "config-manager", "--disable", repo_id))


@pytest.fixture
def dnf_install_rpm():
    installed_rpms = []

    def _install(rpm_name):
        result = subprocess.run(("sudo", "dnf", "-y", "install", rpm_name))
        installed_rpms.append(rpm_name)
        return result

    yield _install
    for rpm_name in installed_rpms:
        subprocess.run(("sudo", "dnf", "-y", "remove", rpm_name))


@pytest.fixture
def create_distribution(
    rpm_metadata_signing_service,
    rpm_repository_factory,
    init_and_sync,
    rpm_publication_factory,
    rpm_distribution_factory,
):
    def _create_distribution(
        repo_config=None,
        has_signing_service=False,
        generate_repo_config=False,
        repo_body=None,
        url=RPM_UNSIGNED_FIXTURE_URL,
        policy="on_demand",
        sync_policy="additive",
    ):
        repo_body = repo_body or {}
        if has_signing_service:
            repo_body["metadata_signing_service"] = rpm_metadata_signing_service.pulp_href
        repo = rpm_repository_factory(**repo_body)
        repo, _ = init_and_sync(repository=repo, url=url, policy=policy, sync_policy=sync_policy)

        pub_body = {"repository": repo.pulp_href}
        if repo_config is not None:
            pub_body["repo_config"] = repo_config
        publication = rpm_publication_factory(**pub_body)

        return rpm_distribution_factory(
            publication=publication.pulp_href, generate_repo_config=generate_repo_config
        )

    return _create_distribution


@pytest.mark.skipif(not dnf_installed, reason="dnf must be installed")
@pytest.mark.parametrize(
    "policy,sync_policy,url",
    [
        ("on_demand", "mirror_complete", RPM_UNSIGNED_FIXTURE_URL),
        ("streamed", "mirror_complete", RPM_UNSIGNED_FIXTURE_URL),
        ("immediate", "mirror_complete", RPM_UNSIGNED_FIXTURE_URL),
        ("on_demand", "mirror_content_only", RPM_UNSIGNED_FIXTURE_URL),
        ("streamed", "mirror_content_only", RPM_UNSIGNED_FIXTURE_URL),
        ("immediate", "mirror_content_only", RPM_UNSIGNED_FIXTURE_URL),
        ("on_demand", "additive", RPM_UNSIGNED_FIXTURE_URL),
        ("streamed", "additive", RPM_UNSIGNED_FIXTURE_URL),
        ("immediate", "additive", RPM_UNSIGNED_FIXTURE_URL),
        # Skipping until we find an XML base repo we can sync from
        # ("immediate", "mirror_content_only", REPO_WITH_XML_BASE_URL),
    ],
)
def test_package_manager_consume(
    policy,
    sync_policy,
    url,
    dnf_config_add_repo,
    dnf_install_rpm,
    create_distribution,
    artifacts_api_client,
    delete_orphans_pre,
):
    """Verify whether package manager can consume content from Pulp."""
    before_sync_artifact_count = artifacts_api_client.list().count

    autopublish = sync_policy != "mirror_complete"
    distribution = create_distribution(
        repo_body={"autopublish": autopublish},
        url=url,
        policy=policy,
        sync_policy=sync_policy,
    )

    before_consumption_artifact_count = artifacts_api_client.list().count
    # sync=mirror_complete creates new Artifacts for the metadata even w/ on_demand & streamed
    # sync!=mirror_complete sets autopublish=True so new Artifacts will also be created
    assert before_consumption_artifact_count > before_sync_artifact_count

    dnf_config_add_repo(distribution)

    rpm_name = "walrus"
    result = dnf_install_rpm(rpm_name)
    assert result.returncode == 0
    rpm = (
        subprocess.run(("rpm", "-q", rpm_name), capture_output=True)
        .stdout.decode("utf-8")
        .strip()
        .split("-")
    )
    assert rpm_name == rpm[0]

    after_consumption_artifact_count = artifacts_api_client.list().count
    if policy == "immediate" or policy == "streamed":
        assert before_consumption_artifact_count == after_consumption_artifact_count
    elif policy == "on_demand":
        assert after_consumption_artifact_count > before_consumption_artifact_count


@pytest.mark.parallel
def test_publish_signed_repo_metadata(
    rpm_metadata_signing_service, create_distribution, dnf_config_add_repo, dnf_install_rpm
):
    """Test if a package manager is able to install packages from a signed repository."""
    if rpm_metadata_signing_service is None:
        pytest.skip("Need a signing service for this test")

    distribution = create_distribution(repo_config={}, has_signing_service=True)
    dnf_config_add_repo(distribution, has_signing_service=True)

    rpm_name = "walrus"
    result = dnf_install_rpm(rpm_name)
    assert result.returncode == 0
    rpm = (
        subprocess.run(("rpm", "-q", rpm_name), capture_output=True)
        .stdout.decode("utf-8")
        .strip()
        .split("-")
    )
    assert rpm_name == rpm[0]


# Test all possible combinations of options made to a publication.
test_options = {
    "repo_config": [{}, {"assumeyes": True, "gpgcheck": 1}],
    "has_signing_service": [True, False],
    "generate_repo_config": [True, False],
}
func_params = itertools.product(*test_options.values())


@pytest.mark.parallel
@pytest.mark.parametrize("repo_config,has_signing_service,generate_repo_config", func_params)
def test_config_dot_repo(
    repo_config,
    has_signing_service,
    generate_repo_config,
    rpm_metadata_signing_service,
    create_distribution,
    http_get,
):
    """Test if the generated config.repo has the right content."""
    if has_signing_service and rpm_metadata_signing_service is None:
        pytest.skip("Need a signing service for this test")
    if not generate_repo_config:
        pytest.skip("Generation of config.repo file was disabled")

    distribution = create_distribution(
        repo_config=repo_config,
        has_signing_service=has_signing_service,
        generate_repo_config=generate_repo_config,
    )
    content = http_get(f"{distribution.base_url}config.repo").decode("utf-8")

    assert f"[{distribution.name}]\n" in content
    assert f"baseurl={distribution.base_url}\n" in content
    assert "gpgcheck=0\n" in content
    assert "repo_gpgcheck=0\n" in content

    if has_signing_service:
        assert f"gpgkey={distribution.base_url}repodata/repomd.xml.key" in content

    if repo_config:
        assert "assumeyes=True\n" in content
        assert "gpgcheck=1\n" in content


@pytest.mark.parallel
def test_repomd_headers(
    create_distribution,
    http_get_headers,
):
    """Test if repomd.xml is returned with Cache-control: no-cache header."""
    distribution = create_distribution(
        repo_config={}, has_signing_service=True, generate_repo_config=True
    )
    assert (
        http_get_headers(f"{distribution.base_url}repodata/repomd.xml").get("Cache-control", "")
        == "no-cache"
    )
    assert (
        not http_get_headers(f"{distribution.base_url}config.repo").get("Cache-control", "")
        == "no-cache"
    )
    assert (
        http_get_headers(f"{distribution.base_url}repodata/repomd.xml.key").get("Cache-control", "")
        == "no-cache"
    )



import createrepo_c as cr
import libcomps
import logging
import os
import tempfile

from django.db import transaction

from pulpcore.plugin.models import PulpTemporaryFile, CreatedResource
from pulpcore.plugin.models import Content
from pulpcore.plugin.util import get_domain

from pulp_rpm.app.comps import strdict_to_dict, dict_digest

from pulp_rpm.app.models import (
    PackageCategory,
    PackageEnvironment,
    PackageGroup,
    PackageLangpacks,
    RpmRepository,
)

log = logging.getLogger(__name__)


def parse_comps_components(comps_file):
    """Parse comps-related components found in the specified file."""
    # created = {"categories": [], "environments": [], "groups": [], "langpack": None}
    created_objects = []
    all_objects = []
    comps = libcomps.Comps()
    curr_domain = get_domain()
    # Read the file and pass the string along because comps.fromxml_f() will only take a
    # path-string that doesn't work on things like S3 storage
    with comps_file.file.open("rb") as comps_uploaded:
        with tempfile.NamedTemporaryFile(dir=".", delete=False) as comps_on_disk:
            comps_on_disk.write(comps_uploaded.read())
            comps_on_disk.flush()
        with tempfile.TemporaryDirectory(dir=".") as tf:
            decompressed_path = os.path.join(tf, "comps.xml")
            cr.decompress_file(comps_on_disk.name, decompressed_path, cr.AUTO_DETECT_COMPRESSION)
            with open(decompressed_path) as f:
                comps.fromxml_str(f.read())

    if comps.langpacks:
        langpack_dict = PackageLangpacks.libcomps_to_dict(comps.langpacks)
        langpack, created = PackageLangpacks.objects.get_or_create(
            matches=strdict_to_dict(comps.langpacks),
            digest=dict_digest(langpack_dict),
            _pulp_domain=curr_domain,
        )
        if created:
            created_objects.append(langpack)
        all_objects.append(langpack)

    if comps.categories:
        for category in comps.categories:
            category_dict = PackageCategory.libcomps_to_dict(category)
            category_dict["digest"] = dict_digest(category_dict)
            category_dict["_pulp_domain"] = curr_domain
            packagecategory, created = PackageCategory.objects.get_or_create(**category_dict)
            if created:
                created_objects.append(packagecategory)
            all_objects.append(packagecategory)

    if comps.environments:
        for environment in comps.environments:
            environment_dict = PackageEnvironment.libcomps_to_dict(environment)
            environment_dict["digest"] = dict_digest(environment_dict)
            environment_dict["_pulp_domain"] = curr_domain
            packageenvironment, created = PackageEnvironment.objects.get_or_create(
                **environment_dict
            )
            if created:
                created_objects.append(packageenvironment)
            all_objects.append(packageenvironment)

    if comps.groups:
        for group in comps.groups:
            group_dict = PackageGroup.libcomps_to_dict(group)
            group_dict["digest"] = dict_digest(group_dict)
            group_dict["_pulp_domain"] = curr_domain
            packagegroup, created = PackageGroup.objects.get_or_create(**group_dict)
            if created:
                created_objects.append(packagegroup)
            all_objects.append(packagegroup)

    return created_objects, all_objects


@transaction.atomic
def upload_comps(tmp_file_id, repo_id=None, replace=False):
    """
    Upload comps.xml file.

    Args:
        tmp_file_id: uploaded comps.xml file.
        repo_id: repository primary key to associate incoming comps-content to.
        replace: if true, replace existing comps-related Content in the specified
            repository with those in the incoming comps.xml file.
    """
    temp_file = PulpTemporaryFile.objects.get(pk=tmp_file_id)
    created, all_objs = parse_comps_components(temp_file)

    for content in all_objs:
        crsrc = CreatedResource(content_object=content)
        crsrc.save()

    if repo_id:
        repository = RpmRepository.objects.get(pk=repo_id)
        if repository:
            all_ids = [obj.content_ptr_id for obj in all_objs]
            with repository.new_version() as new_version:
                if replace:
                    latest = repository.latest_version()
                    rmv_ids = latest.content.filter(
                        pulp_type__in=(
                            PackageCategory.get_pulp_type(),
                            PackageEnvironment.get_pulp_type(),
                            PackageGroup.get_pulp_type(),
                            PackageLangpacks.get_pulp_type(),
                        )
                    )
                    new_version.remove_content(Content.objects.filter(pk__in=rmv_ids))
                new_version.add_content(Content.objects.filter(pk__in=all_ids))



from gettext import gettext as _

import logging

from django.conf import settings
from jsonschema import Draft7Validator
from rest_framework import serializers

from pulpcore.plugin.util import get_domain
from pulpcore.plugin.models import (
    AsciiArmoredDetachedSigningService,
    Remote,
    Publication,
)
from pulpcore.plugin.serializers import (
    RelatedField,
    DetailRelatedField,
    DistributionSerializer,
    PublicationSerializer,
    RemoteSerializer,
    RepositorySerializer,
    RepositorySyncURLSerializer,
    ValidateFieldsMixin,
)

from pulp_rpm.app.constants import (
    ALLOWED_CHECKSUM_ERROR_MSG,
    CHECKSUM_CHOICES,
    SKIP_TYPES,
    SYNC_POLICY_CHOICES,
)
from pulp_rpm.app.models import (
    RpmDistribution,
    RpmRemote,
    RpmRepository,
    RpmPublication,
    UlnRemote,
)
from pulp_rpm.app.schema import COPY_CONFIG_SCHEMA


class RpmRepositorySerializer(RepositorySerializer):
    """
    Serializer for Rpm Repositories.
    """

    autopublish = serializers.BooleanField(
        help_text=_(
            "Whether to automatically create publications for new repository versions, "
            "and update any distributions pointing to this repository."
        ),
        default=False,
        required=False,
    )
    metadata_signing_service = RelatedField(
        help_text="A reference to an associated signing service.",
        view_name="signing-services-detail",
        queryset=AsciiArmoredDetachedSigningService.objects.all(),
        many=False,
        required=False,
        allow_null=True,
    )
    retain_package_versions = serializers.IntegerField(
        help_text=_(
            "The number of versions of each package to keep in the repository; "
            "older versions will be purged. The default is '0', which will disable "
            "this feature and keep all versions of each package."
        ),
        min_value=0,
        required=False,
    )
    metadata_checksum_type = serializers.ChoiceField(
        help_text=_("The checksum type for metadata."),
        choices=CHECKSUM_CHOICES,
        required=False,
        allow_null=True,
    )
    package_checksum_type = serializers.ChoiceField(
        help_text=_("The checksum type for packages."),
        choices=CHECKSUM_CHOICES,
        required=False,
        allow_null=True,
    )
    gpgcheck = serializers.IntegerField(
        max_value=1,
        min_value=0,
        required=False,
        allow_null=True,
        help_text=_(
            "DEPRECATED: An option specifying whether a client should perform "
            "a GPG signature check on packages."
        ),
    )
    repo_gpgcheck = serializers.IntegerField(
        max_value=1,
        min_value=0,
        required=False,
        allow_null=True,
        help_text=_(
            "DEPRECATED: An option specifying whether a client should perform "
            "a GPG signature check on the repodata."
        ),
    )
    sqlite_metadata = serializers.BooleanField(
        default=False,
        required=False,
        help_text=_(
            "DEPRECATED: An option specifying whether Pulp should generate SQLite metadata."
        ),
    )
    repo_config = serializers.JSONField(
        required=False,
        help_text=_("A JSON document describing config.repo file"),
    )

    def validate(self, data):
        """Validate data."""
        for field in ("metadata_checksum_type", "package_checksum_type"):
            if (
                field in data
                and data[field]
                and data[field] not in settings.ALLOWED_CONTENT_CHECKSUMS
            ):
                raise serializers.ValidationError({field: _(ALLOWED_CHECKSUM_ERROR_MSG)})

        validated_data = super().validate(data)
        if (data.get("gpgcheck") or data.get("repo_gpgcheck")) and data.get("repo_config"):
            raise serializers.ValidationError(
                _(
                    "Cannot use gpg options and 'repo_config' options simultaneously. "
                    "The 'gpgcheck' and 'repo_gpgcheck' options are deprecated, please use "
                    "'repo_config' only."
                )
            )
        return validated_data

    def create(self, validated_data):
        """
        Save the repo and handle gpg options

        Args:
            validated_data (dict): A dict of validated data to create the repo

        Returns:
            repo: the created repo
        """
        # gpg options are deprecated in favour of repo_config
        # acting as shim layer between old and new api
        gpgcheck = validated_data.get("gpgcheck")
        repo_gpgcheck = validated_data.get("repo_gpgcheck")

        gpgcheck_options = {}
        if gpgcheck is not None:
            gpgcheck_options["gpgcheck"] = gpgcheck
        if repo_gpgcheck is not None:
            gpgcheck_options["repo_gpgcheck"] = repo_gpgcheck
        if gpgcheck_options.keys():
            logging.getLogger("pulp_rpm.deprecation").info(
                "Support for gpg options will be removed from a future release of pulp_rpm."
            )
        repo_config = (
            gpgcheck_options if gpgcheck_options else validated_data.get("repo_config", {})
        )
        repo = super().create(validated_data)
        repo.repo_config = repo_config
        return repo

    class Meta:
        fields = RepositorySerializer.Meta.fields + (
            "autopublish",
            "metadata_signing_service",
            "retain_package_versions",
            "metadata_checksum_type",
            "package_checksum_type",
            "gpgcheck",
            "repo_gpgcheck",
            "sqlite_metadata",
            "repo_config",
        )
        model = RpmRepository


class RpmBaseRemoteSerializer(RemoteSerializer):
    """
    A common base serializer for multiple RPM based remotes.
    """

    policy = serializers.ChoiceField(
        help_text=_(
            "The policy to use when downloading content. The possible values include: "
            "'immediate', 'on_demand', and 'streamed'. 'immediate' is the default."
        ),
        choices=Remote.POLICY_CHOICES,
        default=Remote.IMMEDIATE,
    )


class RpmRemoteSerializer(RpmBaseRemoteSerializer):
    """
    A Serializer for RpmRemote.
    """

    sles_auth_token = serializers.CharField(
        help_text=_("Authentication token for SLES repositories."),
        required=False,
        allow_null=True,
    )

    class Meta:
        fields = RemoteSerializer.Meta.fields + ("sles_auth_token",)
        model = RpmRemote


class UlnRemoteSerializer(RpmBaseRemoteSerializer):
    """
    A Serializer for UlnRemote.
    """

    username = serializers.CharField(
        help_text=_("Your ULN account username."),
        required=True,
        write_only=True,
    )
    password = serializers.CharField(
        help_text=_("Your ULN account password."),
        required=True,
        write_only=True,
        style={"input_type": "password"},
    )

    url = serializers.CharField(
        help_text=_(
            "The ULN repo URL of the remote content source."
            '"This is "uln://" followed by the channel name. E.g.: "uln://ol7_x86_64_oracle"'
        ),
        required=True,
    )

    uln_server_base_url = serializers.CharField(
        help_text=_(
            "Base URL of the ULN server. If the uln_server_base_url is not provided pulp_rpm will"
            "use the contents of the DEFAULT_ULN_SERVER_BASE_URL setting instead."
        ),
        required=False,
        allow_null=True,
    )

    class Meta:
        fields = RemoteSerializer.Meta.fields + ("uln_server_base_url",)
        model = UlnRemote


class RpmPublicationSerializer(PublicationSerializer):
    """
    A Serializer for RpmPublication.
    """

    metadata_checksum_type = serializers.ChoiceField(
        help_text=_("The checksum type for metadata."),
        choices=CHECKSUM_CHOICES,
        required=False,
    )
    package_checksum_type = serializers.ChoiceField(
        help_text=_("The checksum type for packages."),
        choices=CHECKSUM_CHOICES,
        required=False,
    )
    gpgcheck = serializers.IntegerField(
        max_value=1,
        min_value=0,
        required=False,
        allow_null=True,
        help_text=_(
            "DEPRECATED: An option specifying whether a client should perform "
            "a GPG signature check on packages."
        ),
    )
    repo_gpgcheck = serializers.IntegerField(
        max_value=1,
        min_value=0,
        required=False,
        allow_null=True,
        help_text=_(
            "DEPRECATED: An option specifying whether a client should perform "
            "a GPG signature check on the repodata."
        ),
    )
    sqlite_metadata = serializers.BooleanField(
        default=False,
        required=False,
        help_text=_(
            "DEPRECATED: An option specifying whether Pulp should generate SQLite metadata."
        ),
    )
    repo_config = serializers.JSONField(
        required=False,
        help_text=_("A JSON document describing config.repo file"),
    )

    def validate(self, data):
        """Validate data."""
        if (
            data.get("metadata_checksum_type")
            and data["metadata_checksum_type"] not in settings.ALLOWED_CONTENT_CHECKSUMS
        ) or (
            data.get("package_checksum_type")
            and data["package_checksum_type"] not in settings.ALLOWED_CONTENT_CHECKSUMS
        ):
            raise serializers.ValidationError(_(ALLOWED_CHECKSUM_ERROR_MSG))
        validated_data = super().validate(data)
        if (data.get("gpgcheck") or data.get("repo_gpgcheck")) and data.get("repo_config"):
            raise serializers.ValidationError(
                _(
                    "Cannot use gpg options and 'repo_config' options simultaneously. "
                    "The 'gpgcheck' and 'repo_gpgcheck' options are deprecated, please use "
                    "'repo_config' only."
                )
            )
        return validated_data

    class Meta:
        fields = PublicationSerializer.Meta.fields + (
            "metadata_checksum_type",
            "package_checksum_type",
            "gpgcheck",
            "repo_gpgcheck",
            "sqlite_metadata",
            "repo_config",
        )
        model = RpmPublication


class RpmDistributionSerializer(DistributionSerializer):
    """
    Serializer for RPM Distributions.
    """

    publication = DetailRelatedField(
        required=False,
        help_text=_("Publication to be served"),
        view_name_pattern=r"publications(-.*/.*)?-detail",
        queryset=Publication.objects.exclude(complete=False),
        allow_null=True,
    )
    generate_repo_config = serializers.BooleanField(
        default=False,
        required=False,
        help_text=_("An option specifying whether Pulp should generate *.repo files."),
    )

    class Meta:
        fields = DistributionSerializer.Meta.fields + ("publication", "generate_repo_config")
        model = RpmDistribution


class RpmRepositorySyncURLSerializer(RepositorySyncURLSerializer):
    """
    Serializer for RPM Sync.
    """

    mirror = serializers.BooleanField(
        required=False,
        allow_null=True,
        help_text=_(
            "DEPRECATED: If ``True``, ``sync_policy`` will default to 'mirror_complete' "
            "instead of 'additive'."
        ),
    )
    sync_policy = serializers.ChoiceField(
        help_text=_(
            "Options: 'additive', 'mirror_complete', 'mirror_content_only'. Default: 'additive'. "
            "Modifies how the sync is performed. 'mirror_complete' will clone the original "
            "metadata and create an automatic publication from it, but comes with some "
            "limitations and does not work for certain repositories. 'mirror_content_only' will "
            "change the repository contents to match the remote but the metadata will be "
            "regenerated and will not be bit-for-bit identical. 'additive' will retain the "
            "existing contents of the repository and add the contents of the repository being "
            "synced."
        ),
        choices=SYNC_POLICY_CHOICES,
        required=False,
        allow_null=True,
    )
    skip_types = serializers.ListField(
        help_text=_("List of content types to skip during sync."),
        required=False,
        default=[],
        child=serializers.ChoiceField([(skip_type, skip_type) for skip_type in SKIP_TYPES]),
    )
    optimize = serializers.BooleanField(
        help_text=_("Whether or not to optimize sync."), required=False, default=True
    )

    def validate(self, data):
        """
        Validate sync parameters.
        """
        data = super().validate(data)

        if "mirror" in data and "sync_policy" in data:
            raise serializers.ValidationError(
                _(
                    "Cannot use 'mirror' and 'sync_policy' options simultaneously. The 'mirror' "
                    "option is deprecated, please use 'sync_policy' only."
                )
            )

        return data


class CopySerializer(ValidateFieldsMixin, serializers.Serializer):
    """
    A serializer for Content Copy API.
    """

    config = serializers.JSONField(
        help_text=_("A JSON document describing sources, destinations, and content to be copied"),
    )

    dependency_solving = serializers.BooleanField(
        help_text=_("Also copy dependencies of the content being copied."), default=True
    )

    def validate(self, data):
        """
        Validate that the Serializer contains valid data.

        Make sure the config-JSON matches the config-schema.
        Check for cross-domain references (if domain-enabled).
        """

        def check_domain(domain, href, name):
            # We're doing just a string-check here rather than checking objects
            # because there can be A LOT of objects, and this is happening in the view-layer
            # where we have strictly-limited timescales to work with
            if href and domain not in href:
                raise serializers.ValidationError(
                    _("{} must be a part of the {} domain.").format(name, domain)
                )

        def check_cross_domain_config(cfg):
            """Check that all config-elts are in 'our' domain."""
            # copy-cfg is a list of dictionaries.
            # source_repo_version and dest_repo are required fields.
            # Insure curr-domain exists in src/dest/dest_base_version/content-list hrefs
            curr_domain_name = get_domain().name
            for entry in cfg:
                check_domain(curr_domain_name, entry["source_repo_version"], "dest_repo")
                check_domain(curr_domain_name, entry["dest_repo"], "dest_repo")
                check_domain(
                    curr_domain_name, entry.get("dest_base_version", None), "dest_base_version"
                )
                for content_href in entry.get("content", []):
                    check_domain(curr_domain_name, content_href, "content")

        super().validate(data)
        if "config" in data:
            # Make sure config is valid JSON
            validator = Draft7Validator(COPY_CONFIG_SCHEMA)

            err = []
            for error in sorted(validator.iter_errors(data["config"]), key=str):
                err.append(error.message)
            if err:
                raise serializers.ValidationError(
                    _("Provided copy criteria is invalid:'{}'".format(err))
                )

            if settings.DOMAIN_ENABLED:
                check_cross_domain_config(data["config"])

        return data



from logging import getLogger

import libcomps

from django.db import models

from pulpcore.plugin.models import Content
from pulpcore.plugin.util import get_domain_pk

from pulp_rpm.app.constants import (
    LIBCOMPS_CATEGORY_ATTRS,
    LIBCOMPS_ENVIRONMENT_ATTRS,
    LIBCOMPS_GROUP_ATTRS,
    PULP_CATEGORY_ATTRS,
    PULP_ENVIRONMENT_ATTRS,
    PULP_GROUP_ATTRS,
    PULP_LANGPACKS_ATTRS,
)

from pulp_rpm.app.comps import dict_to_strdict, list_to_idlist, strdict_to_dict

log = getLogger(__name__)


class PackageGroup(Content):
    """
    The "PackageGroup" content type.

    Maps directly to the fields provided by libcomps.
    https://github.com/rpm-software-management/libcomps

    Fields:

        id (Text):
            ID of the group
        default (Bool):
            Flag to identify whether the group is a default
        user_visible (Bool):
            Flag to identify if the group is visible to the user

        display_order (Int):
            Number representing the order of display
        name (Text):
            Name of the group
        description (Text):
            Description of the group
        packages (Text):
            The list of packages in this group
        biarch_only (Bool):
            Flag to identify whether the group is biarch
        desc_by_lang (Text):
            A dictionary of descriptions by language
        name_by_lang (Text):
            A dictionary of names by language
        digest (Text):
            A checksum for the group
    """

    TYPE = "packagegroup"

    # Required metadata
    id = models.TextField()

    default = models.BooleanField(default=False)
    user_visible = models.BooleanField(default=False)

    display_order = models.IntegerField(null=True)
    name = models.TextField()
    description = models.TextField(default="")
    packages = models.JSONField(default=list)

    biarch_only = models.BooleanField(default=False)

    desc_by_lang = models.JSONField(default=dict)
    name_by_lang = models.JSONField(default=dict)

    digest = models.TextField(db_index=True)

    repo_key_fields = ("id",)

    _pulp_domain = models.ForeignKey("core.Domain", default=get_domain_pk, on_delete=models.PROTECT)

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"
        unique_together = ("_pulp_domain", "digest")

    @classmethod
    def pkglist_to_list(cls, value):
        """
        Convert libcomps PkgList to list.

        Args:
            value: a libcomps PkgList

        Returns:
            A list

        """
        package_list = []
        for i in value:
            as_dict = {
                "name": i.name,
                "type": i.type,
                "basearchonly": i.basearchonly,
                "requires": i.requires,
            }
            if as_dict not in package_list:
                package_list.append(as_dict)
        return package_list

    @classmethod
    def list_to_pkglist(cls, lst):
        """
        Convert list of Packages to libcomps PackageList object.

        Args:
            list: a list of Packages

        Returns:
            pkglist: a libcomps PackageList

        """
        pkglist = libcomps.PackageList()
        for pkg in lst:
            lib_pkg = libcomps.Package()
            lib_pkg.name = pkg["name"]
            lib_pkg.type = pkg["type"]
            lib_pkg.basearchonly = bool(pkg["basearchonly"])
            lib_pkg.requires = pkg["requires"]
            pkglist.append(lib_pkg)

        return pkglist

    @classmethod
    def libcomps_to_dict(cls, group):
        """
        Convert libcomps group object to dict for instantiating PackageGroup object.

        Args:
            group(libcomps.group): a RPM/SRPM group to convert

        Returns:
            dict: all data for RPM/SRPM group content creation

        """
        return {
            PULP_GROUP_ATTRS.ID: getattr(group, LIBCOMPS_GROUP_ATTRS.ID),
            PULP_GROUP_ATTRS.DEFAULT: getattr(group, LIBCOMPS_GROUP_ATTRS.DEFAULT),
            PULP_GROUP_ATTRS.USER_VISIBLE: getattr(group, LIBCOMPS_GROUP_ATTRS.USER_VISIBLE),
            PULP_GROUP_ATTRS.DISPLAY_ORDER: getattr(group, LIBCOMPS_GROUP_ATTRS.DISPLAY_ORDER),
            PULP_GROUP_ATTRS.NAME: getattr(group, LIBCOMPS_GROUP_ATTRS.NAME),
            PULP_GROUP_ATTRS.DESCRIPTION: getattr(group, LIBCOMPS_GROUP_ATTRS.DESCRIPTION) or "",
            PULP_GROUP_ATTRS.PACKAGES: cls.pkglist_to_list(
                getattr(group, LIBCOMPS_GROUP_ATTRS.PACKAGES)
            ),
            PULP_GROUP_ATTRS.BIARCH_ONLY: getattr(group, LIBCOMPS_GROUP_ATTRS.BIARCH_ONLY),
            PULP_GROUP_ATTRS.DESC_BY_LANG: strdict_to_dict(
                getattr(group, LIBCOMPS_GROUP_ATTRS.DESC_BY_LANG)
            ),
            PULP_GROUP_ATTRS.NAME_BY_LANG: strdict_to_dict(
                getattr(group, LIBCOMPS_GROUP_ATTRS.NAME_BY_LANG)
            ),
        }

    def pkg_grp_to_libcomps(self):
        """
        Convert PackageGroup object to libcomps Group object.

        Returns:
            group: libcomps.Group object

        """
        group = libcomps.Group()

        group.id = getattr(self, PULP_GROUP_ATTRS.ID)
        group.default = getattr(self, PULP_GROUP_ATTRS.DEFAULT)
        group.uservisible = getattr(self, PULP_GROUP_ATTRS.USER_VISIBLE)
        group.display_order = getattr(self, PULP_GROUP_ATTRS.DISPLAY_ORDER)
        group.name = getattr(self, PULP_GROUP_ATTRS.NAME)
        group.desc = getattr(self, PULP_GROUP_ATTRS.DESCRIPTION)
        group.packages = self.list_to_pkglist(getattr(self, PULP_GROUP_ATTRS.PACKAGES))
        group.biarchonly = getattr(self, PULP_GROUP_ATTRS.BIARCH_ONLY)
        group.desc_by_lang = dict_to_strdict(getattr(self, PULP_GROUP_ATTRS.DESC_BY_LANG))
        group.name_by_lang = dict_to_strdict(getattr(self, PULP_GROUP_ATTRS.NAME_BY_LANG))

        return group


class PackageCategory(Content):
    """
    The "PackageCategory" content type.

    Maps directly to the fields provided by libcomps.
    https://github.com/rpm-software-management/libcomps

    Fields:

        id (Text):
            ID of the category
        name (Text):
            The name of the category
        description (Text):
            The description of the category
        display_order (Int):
            Number representing the order of display
        group_ids (Text):
            A list of group ids
        desc_by_lang (Text):
            A dictionary of descriptions by language
        name_by_lang (Text):
            A dictionary of names by language
        digest (Text):
            A checksum for the category
    """

    TYPE = "packagecategory"

    # Required metadata
    id = models.TextField()

    name = models.TextField()
    description = models.TextField(default="")
    display_order = models.IntegerField(null=True)

    group_ids = models.JSONField(default=list)

    desc_by_lang = models.JSONField(default=dict)
    name_by_lang = models.JSONField(default=dict)

    digest = models.TextField(db_index=True)

    repo_key_fields = ("id",)

    _pulp_domain = models.ForeignKey("core.Domain", default=get_domain_pk, on_delete=models.PROTECT)

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"
        unique_together = ("_pulp_domain", "digest")

    @classmethod
    def grplist_to_lst(cls, value):
        """
        Convert libcomps GrpList to list.

        Args:
            value: a libcomps GrpList

        Returns:
            A list

        """
        grp_list = []
        for i in value:
            grp_list.append({"name": i.name, "default": i.default})
        return grp_list

    @classmethod
    def libcomps_to_dict(cls, category):
        """
        Convert libcomps category object to dict for instantiating PackageCategory object.

        Args:
            category(libcomps.category): a RPM/SRPM category to convert

        Returns:
            dict: all data for RPM/SRPM category content creation

        """
        return {
            PULP_CATEGORY_ATTRS.ID: getattr(category, LIBCOMPS_CATEGORY_ATTRS.ID),
            PULP_CATEGORY_ATTRS.NAME: getattr(category, LIBCOMPS_CATEGORY_ATTRS.NAME),
            PULP_CATEGORY_ATTRS.DESCRIPTION: getattr(category, LIBCOMPS_CATEGORY_ATTRS.DESCRIPTION)
            or "",
            PULP_CATEGORY_ATTRS.DISPLAY_ORDER: getattr(
                category, LIBCOMPS_CATEGORY_ATTRS.DISPLAY_ORDER
            ),
            PULP_CATEGORY_ATTRS.GROUP_IDS: cls.grplist_to_lst(
                getattr(category, LIBCOMPS_CATEGORY_ATTRS.GROUP_IDS)
            ),
            PULP_CATEGORY_ATTRS.DESC_BY_LANG: strdict_to_dict(
                getattr(category, LIBCOMPS_CATEGORY_ATTRS.DESC_BY_LANG)
            ),
            PULP_CATEGORY_ATTRS.NAME_BY_LANG: strdict_to_dict(
                getattr(category, LIBCOMPS_CATEGORY_ATTRS.NAME_BY_LANG)
            ),
        }

    def pkg_cat_to_libcomps(self):
        """
        Convert PackageCategory object to libcomps Category object.

        Returns:
            group: libcomps.Category object

        """
        cat = libcomps.Category()

        cat.id = getattr(self, PULP_CATEGORY_ATTRS.ID)
        cat.name = getattr(self, PULP_CATEGORY_ATTRS.NAME)
        cat.desc = getattr(self, PULP_CATEGORY_ATTRS.DESCRIPTION)
        cat.display_order = getattr(self, PULP_CATEGORY_ATTRS.DISPLAY_ORDER)
        cat.group_ids = list_to_idlist(getattr(self, PULP_CATEGORY_ATTRS.GROUP_IDS))
        cat.desc_by_lang = dict_to_strdict(getattr(self, PULP_CATEGORY_ATTRS.DESC_BY_LANG))
        cat.name_by_lang = dict_to_strdict(getattr(self, PULP_CATEGORY_ATTRS.NAME_BY_LANG))

        return cat


class PackageEnvironment(Content):
    """
    The "PackageEnvironment" content type.

    Maps directly to the fields provided by libcomps.
    https://github.com/rpm-software-management/libcomps

    Fields:

        id (Text):
            ID of the environment
        name (Text):
            The name of the environment
        description (Text):
            The description of the environment
        display_order (Int):
            Number representing the order of display
        group_ids (Text):
            A list of group ids
        option_ids (Text):
            A list of option ids
        desc_by_lang (Text):
            A dictionary of descriptions by language
        name_by_lang (Text):
            A dictionary of names by language
        digest (Text):
            A checksum for the environment
    """

    TYPE = "packageenvironment"

    # Required metadata
    id = models.TextField()

    name = models.TextField()
    description = models.TextField(default="")
    display_order = models.IntegerField(null=True)

    group_ids = models.JSONField(default=list)
    option_ids = models.JSONField(default=list)

    desc_by_lang = models.JSONField(default=dict)
    name_by_lang = models.JSONField(default=dict)

    digest = models.TextField(db_index=True)

    repo_key_fields = ("id",)

    _pulp_domain = models.ForeignKey("core.Domain", default=get_domain_pk, on_delete=models.PROTECT)

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"
        unique_together = ("_pulp_domain", "digest")

    @classmethod
    def grplist_to_lst(cls, value):
        """
        Convert libcomps GrpList to list.

        Args:
            value: a libcomps GrpList

        Returns:
            A list

        """
        grp_list = []
        for i in value:
            grp_list.append({"name": i.name, "default": i.default})
        return grp_list

    @classmethod
    def libcomps_to_dict(cls, environment):
        """
        Convert libcomps environment object to dict for instantiating PackageEnvironment object.

        Args:
            environment(libcomps.environment): a RPM/SRPM environment to convert

        Returns:
            dict: all data for RPM/SRPM environment content creation

        """
        return {
            PULP_ENVIRONMENT_ATTRS.ID: getattr(environment, LIBCOMPS_ENVIRONMENT_ATTRS.ID),
            PULP_ENVIRONMENT_ATTRS.NAME: getattr(environment, LIBCOMPS_ENVIRONMENT_ATTRS.NAME),
            PULP_ENVIRONMENT_ATTRS.DESCRIPTION: getattr(
                environment, LIBCOMPS_ENVIRONMENT_ATTRS.DESCRIPTION
            )
            or "",
            PULP_ENVIRONMENT_ATTRS.DISPLAY_ORDER: getattr(
                environment, LIBCOMPS_ENVIRONMENT_ATTRS.DISPLAY_ORDER
            ),
            PULP_ENVIRONMENT_ATTRS.GROUP_IDS: cls.grplist_to_lst(
                getattr(environment, LIBCOMPS_ENVIRONMENT_ATTRS.GROUP_IDS)
            ),
            PULP_ENVIRONMENT_ATTRS.OPTION_IDS: cls.grplist_to_lst(
                getattr(environment, LIBCOMPS_ENVIRONMENT_ATTRS.OPTION_IDS)
            ),
            PULP_ENVIRONMENT_ATTRS.DESC_BY_LANG: strdict_to_dict(
                getattr(environment, LIBCOMPS_ENVIRONMENT_ATTRS.DESC_BY_LANG)
            ),
            PULP_ENVIRONMENT_ATTRS.NAME_BY_LANG: strdict_to_dict(
                getattr(environment, LIBCOMPS_ENVIRONMENT_ATTRS.NAME_BY_LANG)
            ),
        }

    def pkg_env_to_libcomps(self):
        """
        Convert PackageEnvironment object to libcomps Environment object.

        Returns:
            group: libcomps.Environment object

        """
        env = libcomps.Environment()

        env.id = getattr(self, PULP_ENVIRONMENT_ATTRS.ID)
        env.name = getattr(self, PULP_ENVIRONMENT_ATTRS.NAME)
        env.desc = getattr(self, PULP_ENVIRONMENT_ATTRS.DESCRIPTION)
        env.display_order = getattr(self, PULP_ENVIRONMENT_ATTRS.DISPLAY_ORDER)
        env.group_ids = list_to_idlist(getattr(self, PULP_ENVIRONMENT_ATTRS.GROUP_IDS))
        env.option_ids = list_to_idlist(getattr(self, PULP_ENVIRONMENT_ATTRS.OPTION_IDS))
        env.desc_by_lang = dict_to_strdict(getattr(self, PULP_ENVIRONMENT_ATTRS.DESC_BY_LANG))
        env.name_by_lang = dict_to_strdict(getattr(self, PULP_ENVIRONMENT_ATTRS.NAME_BY_LANG))

        return env


class PackageLangpacks(Content):
    """
    The "PackageLangpacks" content type.

    Maps directly to the fields provided by libcomps.
    https://github.com/rpm-software-management/libcomps

    Fields:

        matches (Dict):
            The langpacks dictionary
    """

    TYPE = "packagelangpacks"

    matches = models.JSONField(default=dict)

    digest = models.TextField(db_index=True)

    _pulp_domain = models.ForeignKey("core.Domain", default=get_domain_pk, on_delete=models.PROTECT)

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"
        unique_together = ("_pulp_domain", "digest")

    @classmethod
    def libcomps_to_dict(cls, langpacks):
        """
        Convert libcomps langpacks object to dict for instantiating PackageLangpacks object.

        Args:
            langpacks(libcomps.langpacks): a RPM/SRPM langpacks to convert

        Returns:
            dict: all data for RPM/SRPM langpacks content creation

        """
        return {PULP_LANGPACKS_ATTRS.MATCHES: strdict_to_dict(langpacks)}



"""Tests that publish rpm plugin repositories."""
import os
import pytest

import collections
from lxml import etree
from django.conf import settings
from xml.etree import ElementTree

import xmltodict
import dictdiffer

from pulp_smash.pulp3.utils import gen_repo, gen_distribution

from pulp_rpm.tests.functional.constants import (
    RPM_ALT_LAYOUT_FIXTURE_URL,
    RPM_COMPLEX_FIXTURE_URL,
    RPM_KICKSTART_FIXTURE_URL,
    RPM_KICKSTART_REPOSITORY_ROOT_CONTENT,
    RPM_LONG_UPDATEINFO_FIXTURE_URL,
    RPM_MODULAR_FIXTURE_URL,
    RPM_NAMESPACES,
    RPM_REFERENCES_UPDATEINFO_URL,
    RPM_RICH_WEAK_FIXTURE_URL,
    RPM_SHA512_FIXTURE_URL,
    RPM_UNSIGNED_FIXTURE_URL,
    SRPM_UNSIGNED_FIXTURE_URL,
)
from pulp_rpm.tests.functional.utils import gen_rpm_remote, read_xml_gz
from pulp_rpm.tests.functional.utils import set_up_module as setUpModule  # noqa:F401

from pulpcore.client.pulp_rpm import RpmRepositorySyncURL, RpmRpmPublication
from pulpcore.client.pulp_rpm.exceptions import ApiException


class TestPublishWithUnsignedRepoSyncedOnDemand:
    @pytest.mark.parallel
    @pytest.mark.parametrize("with_sqlite", [True, False], ids=["with_sqlite", "without_sqlite"])
    def test_sqlite_metadata(
        self,
        with_sqlite,
        rpm_unsigned_repo_on_demand,
        http_get,
        rpm_publication_api,
        gen_object_with_cleanup,
        rpm_distribution_api,
        monitor_task,
    ):
        """Publish repository and validate the updateinfo.

        This Test does the following:

        1. Create a rpm repo and a remote.
        2. Sync the repo with the remote.
        3. Publish with and without sqlite metadata and distribute the repo.
        4. Verify that the sqlite metadata files are/not present when expected.
        """
        publish_data = RpmRpmPublication(
            repository=rpm_unsigned_repo_on_demand.pulp_href, sqlite_metadata=with_sqlite
        )
        publish_response = rpm_publication_api.create(publish_data)
        created_resources = monitor_task(publish_response.task).created_resources
        publication_href = created_resources[0]

        body = gen_distribution(publication=publication_href)
        distribution = gen_object_with_cleanup(rpm_distribution_api, body)

        repomd = ElementTree.fromstring(
            http_get(os.path.join(distribution.base_url, "repodata/repomd.xml"))
        )

        data_xpath = "{{{}}}data".format(RPM_NAMESPACES["metadata/repo"])
        data_elems = [elem for elem in repomd.findall(data_xpath)]

        sqlite_files = [elem for elem in data_elems if elem.get("type").endswith("_db")]

        if with_sqlite:
            assert 3 == len(sqlite_files)

            for db_elem in sqlite_files:
                location_xpath = "{{{}}}location".format(RPM_NAMESPACES["metadata/repo"])
                db_href = db_elem.find(location_xpath).get("href")
                http_get(os.path.join(distribution.base_url, db_href))
        else:
            assert 0 == len(sqlite_files)

    @pytest.mark.parallel
    def test_publish_with_unsupported_checksum_type(
        self, rpm_unsigned_repo_on_demand, rpm_publication_api
    ):
        """
        Sync and try to publish an RPM repository.

        - Sync repository with on_demand policy
        - Try to publish with 'md5' checksum type
        - Publish should fail because 'md5' is not allowed

        This test require disallowed 'MD5' checksum type from ALLOWED_CONTENT_CHECKSUMS settings.
        """
        if "md5" in settings.ALLOWED_CONTENT_CHECKSUMS:
            pytest.skip(
                reason="Cannot verify the expected hasher error if the 'MD5' checksum is allowed."
            )

        publish_data = RpmRpmPublication(
            repository=rpm_unsigned_repo_on_demand.pulp_href, package_checksum_type="md5"
        )
        with pytest.raises(ApiException) as ctx:
            rpm_publication_api.create(publish_data)

        assert "Checksum must be one of the allowed checksum types." in ctx.value.body


class TestPublishWithUnsignedRepoSyncedImmediate:
    @pytest.mark.parallel
    def test_publish_any_repo_version(
        self,
        rpm_unsigned_repo_immediate,
        rpm_publication_api,
        monitor_task,
    ):
        """Test whether a particular repository version can be published."""
        publish_data = RpmRpmPublication(repository=rpm_unsigned_repo_immediate.pulp_href)
        publish_response = rpm_publication_api.create(publish_data)
        created_resources = monitor_task(publish_response.task).created_resources
        publication_href = created_resources[0]

        assert (
            rpm_publication_api.read(publication_href).repository_version
            == rpm_unsigned_repo_immediate.latest_version_href
        )

        non_latest = f"{rpm_unsigned_repo_immediate.pulp_href}versions/0/"

        publish_data.repository_version = non_latest
        publish_data.repository = None
        publish_response = rpm_publication_api.create(publish_data)
        created_resources = monitor_task(publish_response.task).created_resources
        publication_href = created_resources[0]
        publication = rpm_publication_api.read(publication_href)

        assert publication.repository_version == non_latest

        with pytest.raises(ApiException):
            body = {
                "repository": rpm_unsigned_repo_immediate.pulp_href,
                "repository_version": non_latest,
            }
            rpm_publication_api.create(body)

    @pytest.mark.parallel
    def test_validate_no_checksum_tag(
        self,
        rpm_unsigned_repo_immediate,
        rpm_publication_api,
        gen_object_with_cleanup,
        http_get,
        rpm_distribution_api,
        monitor_task,
    ):
        """Sync and publish an RPM repository and verify the checksum.

        TODO: replace with https://github.com/pulp/pulp_rpm/issues/2957
        """
        # 1. Publish and distribute
        publish_data = RpmRpmPublication(repository=rpm_unsigned_repo_immediate.pulp_href)
        publish_response = rpm_publication_api.create(publish_data)
        created_resources = monitor_task(publish_response.task).created_resources
        publication_href = created_resources[0]

        body = gen_distribution(publication=publication_href)
        distribution = gen_object_with_cleanup(rpm_distribution_api, body)

        # 2. check the tag 'sum' is not present in updateinfo.xml
        repomd = ElementTree.fromstring(
            http_get(os.path.join(distribution.base_url, "repodata/repomd.xml"))
        )

        update_xml_url = self._get_updateinfo_xml_path(repomd)
        update_xml_content = http_get(os.path.join(distribution.base_url, update_xml_url))
        update_xml = read_xml_gz(update_xml_content)
        update_info_content = ElementTree.fromstring(update_xml)

        tags = {elem.tag for elem in update_info_content.iter()}
        assert "sum" not in tags, update_info_content

    @staticmethod
    def _get_updateinfo_xml_path(root_elem):
        """Return the path to ``updateinfo.xml.gz``, relative to repository root.

        Given a repomd.xml, this method parses the xml and returns the
        location of updateinfo.xml.gz.
        """
        # <ns0:repomd xmlns:ns0="http://linux.duke.edu/metadata/repo">
        #     <ns0:data type="primary">
        #         <ns0:checksum type="sha256">[…]</ns0:checksum>
        #         <ns0:location href="repodata/[…]-primary.xml.gz" />
        #         …
        #     </ns0:data>
        #     …
        xpath = "{{{}}}data".format(RPM_NAMESPACES["metadata/repo"])
        data_elems = [elem for elem in root_elem.findall(xpath) if elem.get("type") == "updateinfo"]
        xpath = "{{{}}}location".format(RPM_NAMESPACES["metadata/repo"])
        return data_elems[0].find(xpath).get("href")


@pytest.fixture(scope="class")
def assert_created_publication(init_and_sync, rpm_publication_api, monitor_task):
    def _assert_created_publication(url):
        """Sync and publish an RPM repository given a feed URL."""
        repo, _ = init_and_sync(url=url)
        publish_data = RpmRpmPublication(repository=repo.pulp_href)
        publish_response = rpm_publication_api.create(publish_data)
        assert monitor_task(publish_response.task).created_resources != []

    return _assert_created_publication


@pytest.mark.parallel
def test_publish_rpm_rich_weak(assert_created_publication):
    """Sync and publish an RPM rich weak repository."""
    assert_created_publication(RPM_RICH_WEAK_FIXTURE_URL)


@pytest.mark.parallel
def test_publish_rpm_long_updateinfo(assert_created_publication):
    """Sync and publish an RPM long updateinfo repository."""
    assert_created_publication(RPM_LONG_UPDATEINFO_FIXTURE_URL)


@pytest.mark.parallel
def test_publish_rpm_alt_layout(assert_created_publication):
    """Sync and publish an RPM alternate layout repository."""
    assert_created_publication(RPM_ALT_LAYOUT_FIXTURE_URL)


@pytest.mark.parallel
def test_publish_rpm_sha512(assert_created_publication):
    """Sync and publish an RPM SHA256 repository."""
    assert_created_publication(RPM_SHA512_FIXTURE_URL)


@pytest.mark.parallel
def test_publish_srpm(assert_created_publication):
    """Sync and publish an SRPM repository."""
    assert_created_publication(SRPM_UNSIGNED_FIXTURE_URL)


@pytest.mark.parallel
def test_publish_references_update(assert_created_publication):
    """Sync/publish a repo where ``updateinfo.xml`` contains references."""
    # TODO: check the updateinfo metadata to confirm the references are there.
    assert_created_publication(RPM_REFERENCES_UPDATEINFO_URL)


@pytest.mark.parametrize("repo_url", [RPM_COMPLEX_FIXTURE_URL, RPM_MODULAR_FIXTURE_URL])
def test_complex_repo_core_metadata(
    repo_url,
    init_and_sync,
    rpm_publication_api,
    gen_object_with_cleanup,
    http_get,
    rpm_distribution_api,
    monitor_task,
    delete_orphans_pre,
):
    """Test the "complex" fixture that covers more of the metadata cases.

    The standard fixtures have no changelogs and don't cover "ghost" files. The repo
    with the "complex-package" does, and also does a better job of covering rich deps
    and other atypical metadata.
    """

    # create repo and remote
    repo, _ = init_and_sync(url=repo_url, policy="on_demand")

    # publish
    publish_data = RpmRpmPublication(repository=repo.pulp_href)
    publish_response = rpm_publication_api.create(publish_data)
    created_resources = monitor_task(publish_response.task).created_resources
    publication_href = created_resources[0]

    # distribute
    body = gen_distribution(publication=publication_href)
    distribution = gen_object_with_cleanup(rpm_distribution_api, body)

    # Download and parse the metadata.
    original_repomd = ElementTree.fromstring(
        http_get(os.path.join(repo_url, "repodata/repomd.xml"))
    )

    reproduced_repomd = ElementTree.fromstring(
        http_get(os.path.join(distribution.base_url, "repodata/repomd.xml"))
    )

    def get_metadata_content(base_url, repomd_elem, meta_type):
        """Return the text contents of metadata file.

        Provided a url, a repomd root element, and a metadata type, locate the metadata
        file's location href, download it from the provided url, un-gzip it, parse it, and
        return the root element node.

        Don't use this with large repos because it will blow up.
        """
        # <ns0:repomd xmlns:ns0="http://linux.duke.edu/metadata/repo">
        #     <ns0:data type="primary">
        #         <ns0:checksum type="sha256">[…]</ns0:checksum>
        #         <ns0:location href="repodata/[…]-primary.xml.gz" />
        #         …
        #     </ns0:data>
        #     …
        xpath = "{{{}}}data".format(RPM_NAMESPACES["metadata/repo"])
        data_elems = [elem for elem in repomd_elem.findall(xpath) if elem.get("type") == meta_type]
        if not data_elems:
            return None

        xpath = "{{{}}}location".format(RPM_NAMESPACES["metadata/repo"])
        location_href = data_elems[0].find(xpath).get("href")

        return read_xml_gz(http_get(os.path.join(base_url, location_href)))

    # Convert the metadata into a more workable form and then compare.
    for metadata_file in ["primary", "filelists", "other"]:
        original_metadata = get_metadata_content(repo_url, original_repomd, metadata_file)
        generated_metadata = get_metadata_content(
            distribution.base_url, reproduced_repomd, metadata_file
        )

        _compare_xml_metadata_file(original_metadata, generated_metadata, metadata_file)

    # =================

    original_modulemds = get_metadata_content(repo_url, original_repomd, "modules")
    generated_modulemds = get_metadata_content(distribution.base_url, reproduced_repomd, "modules")

    assert bool(original_modulemds) == bool(generated_modulemds)

    if original_modulemds:
        # compare list of modulemd, modulemd-defaults, and modulemd-obsoletes after sorting them
        original_modulemds = sorted(original_modulemds.decode().split("---")[1:])
        generated_modulemds = sorted(generated_modulemds.decode().split("---")[1:])

        assert original_modulemds == generated_modulemds

    # ===================

    # TODO: make this deeper
    original_updateinfo = get_metadata_content(repo_url, original_repomd, "updateinfo")
    generated_updateinfo = get_metadata_content(
        distribution.base_url, reproduced_repomd, "updateinfo"
    )
    assert bool(original_updateinfo) == bool(generated_updateinfo)


def _compare_xml_metadata_file(original_metadata_text, generated_metadata_text, meta_type):
    """Compare two metadata files.

    First convert the metadata into a canonical form. We convert from XML to JSON, and then to
    a dict, and then apply transformations to the dict to standardize the ordering. Then, we
    perform comparisons and observe any differences.
    """
    metadata_block_names = {
        "primary": "metadata",
        "filelists": "filelists",
        "other": "otherdata",
    }
    subsection = metadata_block_names[meta_type]

    # First extract the package entries
    original_metadata = xmltodict.parse(
        original_metadata_text, dict_constructor=collections.OrderedDict
    )[subsection]["package"]
    generated_metadata = xmltodict.parse(
        generated_metadata_text, dict_constructor=collections.OrderedDict
    )[subsection]["package"]

    if not isinstance(original_metadata, list):
        original_metadata = [original_metadata]
    if not isinstance(generated_metadata, list):
        generated_metadata = [generated_metadata]

    # The other transformations are inside the package nodes - they differ by type of metadata
    if meta_type == "primary":
        # location_href gets rewritten by Pulp so we should ignore these differences
        ignore = {"location.@href", "format.file", "time.@file"}
        # we need to make sure all of the requirements are in the same order
        requirement_types = [
            "rpm:suggests",
            "rpm:recommends",
            "rpm:enhances",
            "rpm:provides",
            "rpm:requires",
            "rpm:obsoletes",
            "rpm:conflicts",
            "rpm:supplements",
        ]

        original_metadata = sorted(original_metadata, key=lambda x: x["checksum"]["#text"])
        generated_metadata = sorted(generated_metadata, key=lambda x: x["checksum"]["#text"])
        for md_file in [original_metadata, generated_metadata]:
            for pkg in md_file:
                for req in requirement_types:
                    if pkg["format"].get(req):
                        pkg["format"][req] = sorted(pkg["format"][req])
    elif meta_type == "filelists":
        ignore = {}
        original_metadata = sorted(original_metadata, key=lambda x: x["@pkgid"])
        generated_metadata = sorted(generated_metadata, key=lambda x: x["@pkgid"])
        # make sure the files are all in the same order and type and sort them
        # nodes with a "type" attribute in the XML become OrderedDicts, so we have to convert
        # them to a string representation.
        for md_file in [original_metadata, generated_metadata]:
            for pkg in md_file:
                if pkg.get("file"):
                    files = []
                    for f in pkg["file"]:
                        if isinstance(f, collections.OrderedDict):
                            files.append(
                                "{path} type={type}".format(path=f["@type"], type=f["#text"])
                            )
                        else:
                            files.append(f)

                    pkg["file"] = sorted(files)
    elif meta_type == "other":
        ignore = {}
        original_metadata = sorted(original_metadata, key=lambda x: x["@pkgid"])
        generated_metadata = sorted(generated_metadata, key=lambda x: x["@pkgid"])

        for md_file in [original_metadata, generated_metadata]:
            for pkg in md_file:
                # make sure the changelogs are in the same order
                if pkg.get("changelog"):
                    pkg["changelog"] = sorted(pkg["changelog"], key=lambda x: x["@author"])

    # The metadata dicts should now be consistently ordered. Check for differences.
    for original, generated in zip(original_metadata, generated_metadata):
        diff = dictdiffer.diff(original, generated, ignore=ignore)
        assert list(diff) == [], list(diff)


@pytest.mark.parallel
@pytest.mark.parametrize("mirror", [True, False], ids=["mirror", "standard"])
def test_distribution_tree_metadata_publish(
    mirror,
    gen_object_with_cleanup,
    http_get,
    rpm_repository_api,
    rpm_rpmremote_api,
    rpm_distribution_api,
    monitor_task,
):
    """Test the "complex" fixture that covers more of the metadata cases.

    The standard fixtures have no changelogs and don't cover "ghost" files. The repo
    with the "complex-package" does, and also does a better job of covering rich deps
    and other atypical metadata.
    """
    from configparser import ConfigParser

    # 1. create repo and remote
    repo = gen_object_with_cleanup(rpm_repository_api, gen_repo(autopublish=not mirror))

    body = gen_rpm_remote(RPM_KICKSTART_FIXTURE_URL, policy="on_demand")
    remote = gen_object_with_cleanup(rpm_rpmremote_api, body)

    # 2, 3. Sync and publish
    repository_sync_data = RpmRepositorySyncURL(remote=remote.pulp_href, mirror=mirror)

    sync_response = rpm_repository_api.sync(repo.pulp_href, repository_sync_data)
    created_resources = monitor_task(sync_response.task).created_resources

    publication_href = [r for r in created_resources if "publication" in r][0]

    body = gen_distribution(publication=publication_href)
    distribution = gen_object_with_cleanup(rpm_distribution_api, body)

    # 4. Download and parse the metadata.
    original_treeinfo = http_get(os.path.join(RPM_KICKSTART_FIXTURE_URL, ".treeinfo"))
    generated_treeinfo = http_get(os.path.join(distribution.base_url, ".treeinfo"))

    config = ConfigParser()
    config.optionxform = str  # by default it will cast keys to lower case
    config.read_string(original_treeinfo.decode("utf-8"))
    original_treeinfo = config._sections

    config = ConfigParser()
    config.optionxform = str  # by default it will cast keys to lower case
    config.read_string(generated_treeinfo.decode("utf-8"))
    generated_treeinfo = config._sections

    # 5, 6. Re-arrange the metadata so that it can be compared, and do the comparison.
    # TODO: These really should be in the same order they were in originally.
    # https://pulp.plan.io/issues/9208
    for metadata_dict in [original_treeinfo, generated_treeinfo]:
        metadata_dict["general"]["variants"] = ",".join(
            sorted(metadata_dict["general"]["variants"].split(","))
        )
        metadata_dict["tree"]["variants"] = ",".join(
            sorted(metadata_dict["tree"]["variants"].split(","))
        )

    diff = dictdiffer.diff(original_treeinfo, generated_treeinfo)
    differences = []

    # skip any differences that are "correct" i.e. rewritten "repository" and "packages" paths
    for d in diff:
        (diff_type, diff_name, _, new_value) = (d[0], d[1], d[2][0], d[2][1])
        # ('change', 'variant-Land.packages', ('Packages', 'Land/Packages'))
        if diff_type == "change":
            if diff_name.endswith(".packages") or diff_name.endswith(".repository"):
                # TODO: this is ignoring problems with the generated metadata
                # https://pulp.plan.io/issues/9208
                if "../" not in new_value:
                    continue

        differences.append(d)

    assert differences == [], differences

    # 7. Try downloading the files listed in the .treeinfo metadata, make sure they're
    # actually there.
    for path, checksum in original_treeinfo["checksums"].items():
        if path.startswith("fixtures"):
            # TODO: the .treeinfo metadata is actually wrong for these files, so we can't
            # check them because they won't be there.
            continue

        checksum_type, checksum = checksum.split(":")
        http_get(os.path.join(distribution.base_url, path))


@pytest.fixture
def get_checksum_types(
    init_and_sync,
    http_get,
    rpm_publication_api,
    gen_object_with_cleanup,
    rpm_distribution_api,
    monitor_task,
):
    """Sync and publish an RPM repository."""

    def _get_checksum_types(**kwargs):
        fixture_url = kwargs.get("fixture_url", RPM_UNSIGNED_FIXTURE_URL)
        package_checksum_type = kwargs.get("package_checksum_type")
        metadata_checksum_type = kwargs.get("metadata_checksum_type")
        policy = kwargs.get("policy", "immediate")

        # 1. create repo and remote
        repo, _ = init_and_sync(policy=policy, url=fixture_url, sync_policy="additive")

        # 2. Publish and distribute
        publish_data = RpmRpmPublication(
            repository=repo.pulp_href,
            package_checksum_type=package_checksum_type,
            metadata_checksum_type=metadata_checksum_type,
        )
        publish_response = rpm_publication_api.create(publish_data)
        created_resources = monitor_task(publish_response.task).created_resources
        publication_href = created_resources[0]

        body = gen_distribution(publication=publication_href)
        distribution = gen_object_with_cleanup(rpm_distribution_api, body)

        repomd = ElementTree.fromstring(
            http_get(os.path.join(distribution.base_url, "repodata/repomd.xml"))
        )

        data_xpath = "{{{}}}data".format(RPM_NAMESPACES["metadata/repo"])
        data_elems = [elem for elem in repomd.findall(data_xpath)]

        repomd_checksum_types = {}
        primary_checksum_types = {}
        checksum_xpath = "{{{}}}checksum".format(RPM_NAMESPACES["metadata/repo"])
        for data_elem in data_elems:
            checksum_type = data_elem.find(checksum_xpath).get("type")
            repomd_checksum_types[data_elem.get("type")] = checksum_type
            if data_elem.get("type") == "primary":
                location_xpath = "{{{}}}location".format(RPM_NAMESPACES["metadata/repo"])
                primary_href = data_elem.find(location_xpath).get("href")
                primary = ElementTree.fromstring(
                    read_xml_gz(http_get(os.path.join(distribution.base_url, primary_href)))
                )
                package_checksum_xpath = "{{{}}}checksum".format(RPM_NAMESPACES["metadata/common"])
                package_xpath = "{{{}}}package".format(RPM_NAMESPACES["metadata/common"])
                package_elems = [elem for elem in primary.findall(package_xpath)]
                pkg_checksum_type = package_elems[0].find(package_checksum_xpath).get("type")
                primary_checksum_types[package_elems[0].get("type")] = pkg_checksum_type

        return repomd_checksum_types, primary_checksum_types

    return _get_checksum_types


@pytest.mark.parallel
def test_on_demand_unspecified_checksum_types(get_checksum_types):
    """Sync and publish an RPM repository and verify the checksum types."""
    repomd_checksum_types, primary_checksum_types = get_checksum_types(
        fixture_url=RPM_SHA512_FIXTURE_URL, policy="on_demand"
    )

    for repomd_type, repomd_checksum_type in repomd_checksum_types.items():
        # hack to account for the fact that our md5 and sha512 repos use md5 and sha256
        # checksums for all metadata *except* updateinfo :(
        if not repomd_type == "updateinfo":
            assert repomd_checksum_type == "sha512"

    for package, package_checksum_type in primary_checksum_types.items():
        # since none of the packages in question have sha512 checksums, the
        # checksums they do have will be used instead. In this case, sha512.
        assert package_checksum_type == "sha512"


@pytest.mark.parallel
def test_immediate_unspecified_checksum_types(get_checksum_types):
    """Sync and publish an RPM repository and verify the checksum types."""
    repomd_checksum_types, primary_checksum_types = get_checksum_types(
        fixture_url=RPM_SHA512_FIXTURE_URL, policy="immediate"
    )

    for repomd_type, repomd_checksum_type in repomd_checksum_types.items():
        # hack to account for the fact that our md5 and sha512 repos use md5 and sha256
        # checksums for all metadata *except* updateinfo :(
        if not repomd_type == "updateinfo":
            assert repomd_checksum_type == "sha512"

    for package, package_checksum_type in primary_checksum_types.items():
        assert package_checksum_type == "sha512"


def test_on_demand_specified_package_checksum_type(get_checksum_types, delete_orphans_pre):
    """Sync and publish an RPM repository and verify the checksum types."""
    repomd_checksum_types, primary_checksum_types = get_checksum_types(
        package_checksum_type="sha384", policy="on_demand"
    )

    for repomd_type, repomd_checksum_type in repomd_checksum_types.items():
        assert repomd_checksum_type == "sha256"

    for package, package_checksum_type in primary_checksum_types.items():
        # since none of the packages in question have sha384 checksums, the
        # checksums they do have will be used instead. In this case, sha256.
        assert package_checksum_type == "sha256"


@pytest.mark.parallel
def test_on_demand_specified_metadata_checksum_type(get_checksum_types):
    """Sync and publish an RPM repository and verify the checksum types."""
    repomd_checksum_types, primary_checksum_types = get_checksum_types(
        metadata_checksum_type="sha384", policy="on_demand"
    )

    for repomd_type, repomd_checksum_type in repomd_checksum_types.items():
        assert repomd_checksum_type == "sha384"

    for package, package_checksum_type in primary_checksum_types.items():
        assert package_checksum_type == "sha256"


def test_on_demand_specified_metadata_and_package_checksum_type(
    get_checksum_types, delete_orphans_pre
):
    """Sync and publish an RPM repository and verify the checksum types."""
    repomd_checksum_types, primary_checksum_types = get_checksum_types(
        package_checksum_type="sha224", metadata_checksum_type="sha224", policy="on_demand"
    )

    for repomd_type, repomd_checksum_type in repomd_checksum_types.items():
        assert repomd_checksum_type == "sha224"

    for package, package_checksum_type in primary_checksum_types.items():
        # since none of the packages in question have sha224 checksums, the
        # checksums they do have will be used instead. In this case, sha256.
        assert package_checksum_type == "sha256"


@pytest.mark.parallel
def test_immediate_specified_package_checksum_type(get_checksum_types):
    """Sync and publish an RPM repository and verify the checksum types."""
    repomd_checksum_types, primary_checksum_types = get_checksum_types(
        package_checksum_type="sha384", policy="immediate"
    )

    for repomd_type, repomd_checksum_type in repomd_checksum_types.items():
        assert repomd_checksum_type == "sha256"

    for package, package_checksum_type in primary_checksum_types.items():
        assert package_checksum_type == "sha384"


@pytest.mark.parallel
def test_immediate_specified_metadata_checksum_type(get_checksum_types):
    """Sync and publish an RPM repository and verify the checksum types."""
    repomd_checksum_types, primary_checksum_types = get_checksum_types(
        metadata_checksum_type="sha384", policy="immediate"
    )

    for repomd_type, repomd_checksum_type in repomd_checksum_types.items():
        assert repomd_checksum_type == "sha384"

    for package, package_checksum_type in primary_checksum_types.items():
        assert package_checksum_type == "sha256"


@pytest.mark.parallel
def test_immediate_specified_metadata_and_package_checksum_type(get_checksum_types):
    """Sync and publish an RPM repository and verify the checksum types."""
    repomd_checksum_types, primary_checksum_types = get_checksum_types(
        package_checksum_type="sha224", metadata_checksum_type="sha224", policy="immediate"
    )

    for repomd_type, repomd_checksum_type in repomd_checksum_types.items():
        assert repomd_checksum_type == "sha224"

    for package, package_checksum_type in primary_checksum_types.items():
        assert package_checksum_type == "sha224"


@pytest.mark.parallel
def test_directory_layout_distribute_with_modules(generate_distribution, http_get):
    """Ensure no more files or folders are present when distribute repository with modules."""
    distribution = generate_distribution(RPM_MODULAR_FIXTURE_URL)
    parser = etree.XMLParser(recover=True)
    repository = etree.fromstring(http_get(distribution), parser=parser)
    # Get links from repository HTML
    # Each link is an item (file or directory) in repository root
    repository_root_items = []
    for elem in repository.iter():
        if elem.tag == "a" and not elem.text.startswith(".."):  # skip parent-dir if present
            repository_root_items.append(elem.attrib["href"])

    # Check if 'Packages' and 'repodata' are present
    # Trailing '/' is present for easier check
    assert "Packages/" in repository_root_items
    assert "repodata/" in repository_root_items
    # Only these three items should be present
    assert len(repository_root_items) == 2


@pytest.mark.parallel
def test_directory_layout_distribute_with_treeinfo(generate_distribution, http_get):
    """Ensure no more files or folders are present when distribute repository with treeinfo."""
    distribution = generate_distribution(RPM_KICKSTART_FIXTURE_URL)
    parser = etree.XMLParser(recover=True)
    repository = etree.fromstring(http_get(distribution), parser=parser)
    # Get links from repository HTML
    # Each link is an item (file or directory) in repository root
    repository_root_items = []
    for elem in repository.iter():
        if elem.tag == "a" and not elem.text.startswith(".."):  # skip parent-dir if present
            repository_root_items.append(elem.attrib["href"])
    # Check if all treeinfo related directories are present
    # Trailing '/' is present for easier check
    for directory in RPM_KICKSTART_REPOSITORY_ROOT_CONTENT:
        assert directory in repository_root_items

    assert "repodata/" in repository_root_items
    # assert how many items are present altogether
    # here is '+1' for 'repodata'
    assert len(repository_root_items) == len(RPM_KICKSTART_REPOSITORY_ROOT_CONTENT) + 1


@pytest.fixture(scope="class")
def generate_distribution(
    init_and_sync,
    gen_object_with_cleanup,
    rpm_distribution_api,
    rpm_publication_api,
    monitor_task,
):
    def _generate_distribution(url=None):
        """Sync and publish an RPM repository.

        - create repository
        - create remote
        - sync the remote
        - create publication
        - create distribution

        Args:
            url(string):
                Optional URL of repository that should be use as a remote

        Returns (string):
            RPM distribution base_url.
        """
        repo, _ = init_and_sync(url=url)

        publish_data = RpmRpmPublication(repository=repo.pulp_href)
        publish_response = rpm_publication_api.create(publish_data)
        created_resources = monitor_task(publish_response.task).created_resources
        publication_href = created_resources[0]

        body = gen_distribution(publication=publication_href)
        distribution = gen_object_with_cleanup(rpm_distribution_api, body)

        return distribution.to_dict()["base_url"]

    return _generate_distribution



import copy
from gettext import gettext as _
import json

import createrepo_c
from django.db import IntegrityError, transaction
from rest_framework import serializers

from pulpcore.plugin.serializers import (
    ModelSerializer,
    NoArtifactContentUploadSerializer,
)

from pulp_rpm.app.advisory import hash_update_record
from pulp_rpm.app.fields import (
    UpdateCollectionPackagesField,
    UpdateReferenceField,
)

from pulp_rpm.app.constants import (
    CR_UPDATE_REFERENCE_ATTRS,
    PULP_UPDATE_COLLECTION_ATTRS,
    PULP_UPDATE_RECORD_ATTRS,
    PULP_UPDATE_REFERENCE_ATTRS,
)

from pulp_rpm.app.models import (
    UpdateCollection,
    UpdateCollectionPackage,
    UpdateRecord,
    UpdateReference,
)


class UpdateCollectionSerializer(ModelSerializer):
    """
    A Serializer for UpdateCollection.
    """

    name = serializers.CharField(help_text=_("Collection name."), allow_blank=True, allow_null=True)

    shortname = serializers.CharField(
        help_text=_("Collection short name."), allow_blank=True, allow_null=True
    )

    module = serializers.JSONField(help_text=_("Collection modular NSVCA."), allow_null=True)

    packages = UpdateCollectionPackagesField(
        source="*", read_only=True, help_text=_("List of packages")
    )

    class Meta:
        fields = ("name", "shortname", "module", "packages")
        model = UpdateCollection


class UpdateRecordSerializer(NoArtifactContentUploadSerializer):
    """
    A Serializer for UpdateRecord.
    """

    id = serializers.CharField(
        help_text=_("Update id (short update name, e.g. RHEA-2013:1777)"),
        read_only=True,
    )
    updated_date = serializers.CharField(
        help_text=_("Date when the update was updated (e.g. '2013-12-02 00:00:00')"),
        read_only=True,
    )

    description = serializers.CharField(
        help_text=_("Update description"), allow_blank=True, read_only=True
    )
    issued_date = serializers.CharField(
        help_text=_("Date when the update was issued (e.g. '2013-12-02 00:00:00')"),
        read_only=True,
    )
    fromstr = serializers.CharField(
        help_text=_("Source of the update (e.g. security@redhat.com)"),
        allow_blank=True,
        read_only=True,
    )
    status = serializers.CharField(
        help_text=_("Update status ('final', ...)"), allow_blank=True, read_only=True
    )
    title = serializers.CharField(help_text=_("Update name"), allow_blank=True, read_only=True)
    summary = serializers.CharField(help_text=_("Short summary"), allow_blank=True, read_only=True)
    version = serializers.CharField(
        help_text=_("Update version (probably always an integer number)"),
        allow_blank=True,
        read_only=True,
    )

    type = serializers.CharField(
        help_text=_("Update type ('enhancement', 'bugfix', ...)"),
        allow_blank=True,
        read_only=True,
    )
    severity = serializers.CharField(help_text=_("Severity"), allow_blank=True, read_only=True)
    solution = serializers.CharField(help_text=_("Solution"), allow_blank=True, read_only=True)
    release = serializers.CharField(help_text=_("Update release"), allow_blank=True, read_only=True)
    rights = serializers.CharField(help_text=_("Copyrights"), allow_blank=True, read_only=True)
    pushcount = serializers.CharField(help_text=_("Push count"), allow_blank=True, read_only=True)
    reboot_suggested = serializers.BooleanField(help_text=_("Reboot suggested"), read_only=True)
    pkglist = UpdateCollectionSerializer(
        source="collections", read_only=True, many=True, help_text=_("List of packages")
    )
    references = UpdateReferenceField(
        source="pk", read_only=True, help_text=_("List of references")
    )

    def create(self, validated_data):
        """
        Create UpdateRecord and its subclasses from JSON file.

        Returns:
            UpdateRecord instance

        """
        references = validated_data.pop("references", [])
        pkglist = validated_data.pop("pkglist", [])

        # detach any specified repository; attach once the advisory exists
        repository = validated_data.pop("repository", None)

        update_collection_packages_to_save = list()
        update_references_to_save = list()
        with transaction.atomic():
            try:
                # This persists an advisory with an empty digest
                update_record = super().create(validated_data)
            except IntegrityError:
                # At this point, the advisory has an empty digest. If we hit this,
                # it means a previous advisory-create failed to clean up after itself.
                raise serializers.ValidationError("Advisory already exists in Pulp.")

            for collection in pkglist:
                new_coll = copy.deepcopy(collection)
                packages = new_coll.pop("packages", [])
                new_coll[PULP_UPDATE_COLLECTION_ATTRS.SHORTNAME] = new_coll.pop("short", "")
                coll = UpdateCollection(**new_coll)
                coll.update_record = update_record
                coll.save()
                for package in packages:
                    pkg = UpdateCollectionPackage(**package)
                    try:
                        pkg.sum_type = createrepo_c.checksum_type(pkg.sum_type)
                    except TypeError:
                        raise TypeError(f'"{pkg.sum_type}" is not supported.')
                    pkg.update_collection = coll
                    update_collection_packages_to_save.append(pkg)
            for reference in references:
                new_ref = dict()
                new_ref[PULP_UPDATE_REFERENCE_ATTRS.HREF] = reference.get(
                    CR_UPDATE_REFERENCE_ATTRS.HREF, ""
                )
                new_ref[PULP_UPDATE_REFERENCE_ATTRS.ID] = reference.get(
                    CR_UPDATE_REFERENCE_ATTRS.ID, ""
                )
                new_ref[PULP_UPDATE_REFERENCE_ATTRS.TITLE] = reference.get(
                    CR_UPDATE_REFERENCE_ATTRS.TITLE, ""
                )
                new_ref[PULP_UPDATE_REFERENCE_ATTRS.TYPE] = reference.get(
                    CR_UPDATE_REFERENCE_ATTRS.TYPE, ""
                )
                ref = UpdateReference(**new_ref)
                ref.update_record = update_record
                update_references_to_save.append(ref)

            if update_collection_packages_to_save:
                UpdateCollectionPackage.objects.bulk_create(update_collection_packages_to_save)
            if update_references_to_save:
                UpdateReference.objects.bulk_create(update_references_to_save)

            cr_update_record = update_record.to_createrepo_c()
            update_record.digest = hash_update_record(cr_update_record)

            # The advisory now has a digest - *if* this works
            update_record.save()

        # create new repo version with uploaded advisory
        if repository:
            repository.cast()
            content_to_add = self.Meta.model.objects.filter(pk=update_record.pk)
            with repository.new_version() as new_version:
                new_version.add_content(content_to_add)

        return update_record

    def deferred_validate(self, data):
        """
        Read a file for a JSON data and validate a UpdateRecord data.

        Also change few fields to match Pulp internals if exists as this is usually handle by
        createrepo_c which is not used here.
        """
        data = super().deferred_validate(data)

        update_record_data = dict()
        try:
            update_record_data.update(json.loads(data["file"].read()))
        except UnicodeDecodeError:
            raise serializers.ValidationError("JSON file is expected")
        update_record_data.update(data)

        update_record_data[PULP_UPDATE_RECORD_ATTRS.FROMSTR] = update_record_data.pop(
            "from", update_record_data.get(PULP_UPDATE_RECORD_ATTRS.FROMSTR, "")
        )
        update_record_data[PULP_UPDATE_RECORD_ATTRS.ISSUED_DATE] = update_record_data.pop(
            "issued", update_record_data.get(PULP_UPDATE_RECORD_ATTRS.ISSUED_DATE, "")
        )
        update_record_data[PULP_UPDATE_RECORD_ATTRS.UPDATED_DATE] = update_record_data.pop(
            "updated", update_record_data.get(PULP_UPDATE_RECORD_ATTRS.UPDATED_DATE, "")
        )

        if (
            not update_record_data.get(PULP_UPDATE_RECORD_ATTRS.ID)
            or not update_record_data.get(PULP_UPDATE_RECORD_ATTRS.UPDATED_DATE)
            or not update_record_data.get(PULP_UPDATE_RECORD_ATTRS.ISSUED_DATE)
        ):
            raise serializers.ValidationError(
                "All '{}', '{}' and '{}' must be specified.".format(
                    PULP_UPDATE_RECORD_ATTRS.ID,
                    PULP_UPDATE_RECORD_ATTRS.UPDATED_DATE,
                    PULP_UPDATE_RECORD_ATTRS.ISSUED_DATE,
                )
            )

        return update_record_data

    def retrieve(self, validated_data):
        # This feature cannot be implemented until we can reliably calculate the digest of the
        # advisory before saving anything to the database. By current standards, we can't.
        return None

    class Meta:
        fields = NoArtifactContentUploadSerializer.Meta.fields + (
            "id",
            "updated_date",
            "description",
            "issued_date",
            "fromstr",
            "status",
            "title",
            "summary",
            "version",
            "type",
            "severity",
            "solution",
            "release",
            "rights",
            "pushcount",
            "pkglist",
            "references",
            "reboot_suggested",
        )
        model = UpdateRecord


class MinimalUpdateRecordSerializer(UpdateRecordSerializer):
    """
    A minimal serializer for RPM update records.
    """

    class Meta:
        fields = NoArtifactContentUploadSerializer.Meta.fields + (
            "id",
            "title",
            "severity",
            "type",
        )
        model = UpdateRecord



# Generated by Django 4.2.1 on 2023-05-15 16:51

from django.db import migrations, models
import django.db.models.deletion
import pulpcore.app.util


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0106_alter_artifactdistribution_distribution_ptr_and_more'),
        ('rpm', '0050_alter_addon_pulp_id_alter_checksum_pulp_id_and_more'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='distributiontree',
            unique_together=set(),
        ),
        migrations.AlterUniqueTogether(
            name='modulemd',
            unique_together=set(),
        ),
        migrations.AlterUniqueTogether(
            name='modulemdobsolete',
            unique_together=set(),
        ),
        migrations.AlterUniqueTogether(
            name='package',
            unique_together=set(),
        ),
        migrations.AlterUniqueTogether(
            name='repometadatafile',
            unique_together=set(),
        ),
        migrations.AddField(
            model_name='distributiontree',
            name='_pulp_domain',
            field=models.ForeignKey(default=pulpcore.app.util.get_domain_pk, on_delete=django.db.models.deletion.PROTECT, to='core.domain'),
        ),
        migrations.AddField(
            model_name='modulemd',
            name='_pulp_domain',
            field=models.ForeignKey(default=pulpcore.app.util.get_domain_pk, on_delete=django.db.models.deletion.PROTECT, to='core.domain'),
        ),
        migrations.AddField(
            model_name='modulemddefaults',
            name='_pulp_domain',
            field=models.ForeignKey(default=pulpcore.app.util.get_domain_pk, on_delete=django.db.models.deletion.PROTECT, to='core.domain'),
        ),
        migrations.AddField(
            model_name='modulemdobsolete',
            name='_pulp_domain',
            field=models.ForeignKey(default=pulpcore.app.util.get_domain_pk, on_delete=django.db.models.deletion.PROTECT, to='core.domain'),
        ),
        migrations.AddField(
            model_name='package',
            name='_pulp_domain',
            field=models.ForeignKey(default=pulpcore.app.util.get_domain_pk, on_delete=django.db.models.deletion.PROTECT, to='core.domain'),
        ),
        migrations.AddField(
            model_name='packagecategory',
            name='_pulp_domain',
            field=models.ForeignKey(default=pulpcore.app.util.get_domain_pk, on_delete=django.db.models.deletion.PROTECT, to='core.domain'),
        ),
        migrations.AddField(
            model_name='packageenvironment',
            name='_pulp_domain',
            field=models.ForeignKey(default=pulpcore.app.util.get_domain_pk, on_delete=django.db.models.deletion.PROTECT, to='core.domain'),
        ),
        migrations.AddField(
            model_name='packagegroup',
            name='_pulp_domain',
            field=models.ForeignKey(default=pulpcore.app.util.get_domain_pk, on_delete=django.db.models.deletion.PROTECT, to='core.domain'),
        ),
        migrations.AddField(
            model_name='packagelangpacks',
            name='_pulp_domain',
            field=models.ForeignKey(default=pulpcore.app.util.get_domain_pk, on_delete=django.db.models.deletion.PROTECT, to='core.domain'),
        ),
        migrations.AddField(
            model_name='repometadatafile',
            name='_pulp_domain',
            field=models.ForeignKey(default=pulpcore.app.util.get_domain_pk, on_delete=django.db.models.deletion.PROTECT, to='core.domain'),
        ),
        migrations.AddField(
            model_name='updaterecord',
            name='_pulp_domain',
            field=models.ForeignKey(default=pulpcore.app.util.get_domain_pk, on_delete=django.db.models.deletion.PROTECT, to='core.domain'),
        ),
        migrations.AlterField(
            model_name='modulemddefaults',
            name='digest',
            field=models.TextField(db_index=True),
        ),
        migrations.AlterField(
            model_name='package',
            name='pkgId',
            field=models.TextField(db_index=True),
        ),
        migrations.AlterField(
            model_name='packagecategory',
            name='digest',
            field=models.TextField(db_index=True),
        ),
        migrations.AlterField(
            model_name='packageenvironment',
            name='digest',
            field=models.TextField(db_index=True),
        ),
        migrations.AlterField(
            model_name='packagegroup',
            name='digest',
            field=models.TextField(db_index=True),
        ),
        migrations.AlterField(
            model_name='packagelangpacks',
            name='digest',
            field=models.TextField(db_index=True),
        ),
        migrations.AlterField(
            model_name='updaterecord',
            name='digest',
            field=models.TextField(db_index=True),
        ),
        migrations.AlterUniqueTogether(
            name='distributiontree',
            unique_together={('_pulp_domain', 'digest')},
        ),
        migrations.AlterUniqueTogether(
            name='modulemd',
            unique_together={('_pulp_domain', 'name', 'stream', 'version', 'context', 'arch')},
        ),
        migrations.AlterUniqueTogether(
            name='modulemddefaults',
            unique_together={('_pulp_domain', 'digest')},
        ),
        migrations.AlterUniqueTogether(
            name='modulemdobsolete',
            unique_together={('_pulp_domain', 'modified', 'module_name', 'module_stream')},
        ),
        migrations.AlterUniqueTogether(
            name='package',
            unique_together={('_pulp_domain', 'name', 'epoch', 'version', 'release', 'arch', 'checksum_type', 'pkgId')},
        ),
        migrations.AlterUniqueTogether(
            name='packagecategory',
            unique_together={('_pulp_domain', 'digest')},
        ),
        migrations.AlterUniqueTogether(
            name='packageenvironment',
            unique_together={('_pulp_domain', 'digest')},
        ),
        migrations.AlterUniqueTogether(
            name='packagegroup',
            unique_together={('_pulp_domain', 'digest')},
        ),
        migrations.AlterUniqueTogether(
            name='packagelangpacks',
            unique_together={('_pulp_domain', 'digest')},
        ),
        migrations.AlterUniqueTogether(
            name='repometadatafile',
            unique_together={('_pulp_domain', 'data_type', 'checksum', 'relative_path')},
        ),
        migrations.AlterUniqueTogether(
            name='updaterecord',
            unique_together={('_pulp_domain', 'digest')},
        ),
    ]



from gettext import gettext as _

from rest_framework import serializers

from pulpcore.plugin.models import Repository
from pulpcore.plugin.serializers import DetailRelatedField
from pulpcore.plugin.serializers import NoArtifactContentSerializer

from pulp_rpm.app.models import (
    PackageCategory,
    PackageEnvironment,
    PackageGroup,
    PackageLangpacks,
)


class PackageGroupSerializer(NoArtifactContentSerializer):
    """
    PackageGroup serializer.
    """

    id = serializers.CharField(
        help_text=_("PackageGroup id."),
    )
    default = serializers.BooleanField(help_text=_("PackageGroup default."), required=False)
    user_visible = serializers.BooleanField(
        help_text=_("PackageGroup user visibility."), required=False
    )
    display_order = serializers.IntegerField(
        help_text=_("PackageGroup display order."), allow_null=True
    )
    name = serializers.CharField(help_text=_("PackageGroup name."), allow_blank=True)
    description = serializers.CharField(help_text=_("PackageGroup description."), allow_blank=True)
    packages = serializers.JSONField(help_text=_("PackageGroup package list."), allow_null=True)
    biarch_only = serializers.BooleanField(help_text=_("PackageGroup biarch only."), required=False)
    desc_by_lang = serializers.JSONField(
        help_text=_("PackageGroup description by language."), allow_null=True
    )
    name_by_lang = serializers.JSONField(
        help_text=_("PackageGroup name by language."), allow_null=True
    )
    digest = serializers.CharField(
        help_text=_("PackageGroup digest."),
    )

    class Meta:
        fields = NoArtifactContentSerializer.Meta.fields + (
            "id",
            "default",
            "user_visible",
            "display_order",
            "name",
            "description",
            "packages",
            "biarch_only",
            "desc_by_lang",
            "name_by_lang",
            "digest",
        )
        model = PackageGroup


class PackageCategorySerializer(NoArtifactContentSerializer):
    """
    PackageCategory serializer.
    """

    id = serializers.CharField(
        help_text=_("Category id."),
    )
    name = serializers.CharField(help_text=_("Category name."), allow_blank=True)
    description = serializers.CharField(help_text=_("Category description."), allow_blank=True)
    display_order = serializers.IntegerField(
        help_text=_("Category display order."), allow_null=True
    )
    group_ids = serializers.JSONField(help_text=_("Category group list."), allow_null=True)
    desc_by_lang = serializers.JSONField(
        help_text=_("Category description by language."), allow_null=True
    )
    name_by_lang = serializers.JSONField(help_text=_("Category name by language."), allow_null=True)
    digest = serializers.CharField(
        help_text=_("Category digest."),
    )

    class Meta:
        fields = NoArtifactContentSerializer.Meta.fields + (
            "id",
            "name",
            "description",
            "display_order",
            "group_ids",
            "desc_by_lang",
            "name_by_lang",
            "digest",
        )
        model = PackageCategory


class PackageEnvironmentSerializer(NoArtifactContentSerializer):
    """
    PackageEnvironment serializer.
    """

    id = serializers.CharField(
        help_text=_("Environment id."),
    )
    name = serializers.CharField(help_text=_("Environment name."), allow_blank=True)
    description = serializers.CharField(help_text=_("Environment description."), allow_blank=True)
    display_order = serializers.IntegerField(
        help_text=_("Environment display order."), allow_null=True
    )
    group_ids = serializers.JSONField(help_text=_("Environment group list."), allow_null=True)
    option_ids = serializers.JSONField(help_text=_("Environment option ids"), allow_null=True)
    desc_by_lang = serializers.JSONField(
        help_text=_("Environment description by language."), allow_null=True
    )
    name_by_lang = serializers.JSONField(
        help_text=_("Environment name by language."), allow_null=True
    )
    digest = serializers.CharField(help_text=_("Environment digest."))

    class Meta:
        fields = NoArtifactContentSerializer.Meta.fields + (
            "id",
            "name",
            "description",
            "display_order",
            "group_ids",
            "option_ids",
            "desc_by_lang",
            "name_by_lang",
            "digest",
        )
        model = PackageEnvironment


class PackageLangpacksSerializer(NoArtifactContentSerializer):
    """
    PackageLangpacks serializer.
    """

    matches = serializers.JSONField(help_text=_("Langpacks matches."), allow_null=True)
    digest = serializers.CharField(help_text=_("Langpacks digest."), allow_null=True)

    class Meta:
        fields = NoArtifactContentSerializer.Meta.fields + ("matches", "digest")
        model = PackageLangpacks


class CompsXmlSerializer(serializers.Serializer):
    """
    A serializer for comps.xml Upload API.
    """

    file = serializers.FileField(
        help_text=_(
            "Full path of a comps.xml file that may be parsed into comps.xml Content units."
        ),
        required=True,
    )
    repository = DetailRelatedField(
        help_text=_(
            "URI of an RPM repository the comps.xml content units should be associated to."
        ),
        required=False,
        write_only=True,
        view_name_pattern=r"repositories(-.*/.*)-detail",
        queryset=Repository.objects.all(),
    )

    replace = serializers.BooleanField(
        help_text=_(
            "If true, incoming comps.xml replaces existing comps-related ContentUnits in the "
            "specified repository."
        ),
        required=False,
        write_only=True,
    )

    class Meta:
        fields = ("file", "repository", "replace")



# Generated by Django 2.2.10 on 2020-02-19 14:43


import json
import yaml

from django.db import migrations, transaction
import django.contrib.postgres.fields.jsonb


def unflatten_json(apps, schema_editor):
    # re: https://pulp.plan.io/issues/6191
    with transaction.atomic():
        Modulemd = apps.get_model("rpm", "Modulemd")
        for module in Modulemd.objects.all().only("artifacts", "dependencies", "_artifacts"):
            a = module._artifacts.first()
            snippet = a.file.read().decode()
            module_dict = yaml.safe_load(snippet)
            a.file.close()
            module.artifacts = json.loads(module.artifacts)
            module.dependencies = module_dict["data"]["dependencies"]
            module.save()

        ModulemdDefaults = apps.get_model("rpm", "ModulemdDefaults")
        for mod_defs in ModulemdDefaults.objects.all().only("profiles"):
            mod_defs.profiles = json.loads(mod_defs.profiles)
            mod_defs.save()


def replace_emptystring_with_null(apps, schema_editor):
    # An empty string can't be converted to JSON, so use a null instead
    with transaction.atomic():
        UpdateCollection = apps.get_model("rpm", "UpdateCollection")
        for coll in UpdateCollection.objects.all().only("module"):
            if coll.module == "":
                coll.module = "null"  # de-serializes to None,
                coll.save()


class Migration(migrations.Migration):

    dependencies = [
        ("rpm", "0002_updaterecord_reboot_suggested"),
    ]

    operations = [
        # Bugfixes
        migrations.RunPython(unflatten_json),
        # In-place migrate JSON stored in text field to a JSONField
        migrations.RunPython(replace_emptystring_with_null),
        migrations.AlterField(
            model_name="updatecollection",
            name="module",
            field=django.contrib.postgres.fields.jsonb.JSONField(null=True),
        ),
    ]



# coding=utf-8
"""Tests that sync rpm plugin repositories."""
import pytest


from pulpcore.client.pulp_rpm import (
    RpmRepositorySyncURL,
)


@pytest.fixture
def setup_autopublish(rpm_repository_factory, rpm_rpmremote_factory, rpm_distribution_factory):
    """Create remote, repo, publish settings, and distribution."""
    remote = rpm_rpmremote_factory()
    repo = rpm_repository_factory(autopublish=True, sqlite_metadata=True)
    distribution = rpm_distribution_factory(repository=repo.pulp_href)

    return repo, remote, distribution


@pytest.mark.parallel
def test_01_sync(setup_autopublish, rpm_repository_api, rpm_publication_api, monitor_task):
    """Assert that syncing the repository triggers auto-publish and auto-distribution."""
    repo, remote, distribution = setup_autopublish
    assert rpm_publication_api.list(repository=repo.pulp_href).count == 0
    assert distribution.publication is None

    # Sync the repository.
    repository_sync_data = RpmRepositorySyncURL(remote=remote.pulp_href)
    sync_response = rpm_repository_api.sync(repo.pulp_href, repository_sync_data)
    task = monitor_task(sync_response.task)

    # Check that all the appropriate resources were created
    assert len(task.created_resources) > 1
    publications = rpm_publication_api.list(repository=repo.pulp_href)
    assert publications.count == 1

    # Check that the publish settings were used
    publication = publications.results[0]
    assert publication.sqlite_metadata is True

    # Sync the repository again. Since there should be no new repository version, there
    # should be no new publications or distributions either.
    sync_response = rpm_repository_api.sync(repo.pulp_href, repository_sync_data)
    task = monitor_task(sync_response.task)

    assert len(task.created_resources) == 0
    assert rpm_publication_api.list(repository=repo.pulp_href).count == 1


@pytest.mark.parallel
def test_02_modify(
    setup_autopublish, rpm_repository_api, rpm_package_api, rpm_publication_api, monitor_task
):
    """Assert that modifying the repository triggers auto-publish and auto-distribution."""
    repo, remote, distribution = setup_autopublish
    assert rpm_publication_api.list(repository=repo.pulp_href).count == 0
    assert distribution.publication is None

    # Modify the repository by adding a content unit
    content = rpm_package_api.list().results[0].pulp_href

    modify_response = rpm_repository_api.modify(repo.pulp_href, {"add_content_units": [content]})
    task = monitor_task(modify_response.task)

    # Check that all the appropriate resources were created
    assert len(task.created_resources) > 1
    publications = rpm_publication_api.list(repository=repo.pulp_href)
    assert publications.count == 1

    # Check that the publish settings were used
    publication = publications.results[0]
    assert publication.sqlite_metadata is True



"""Tests for Pulp's characters encoding."""
import pytest
import uuid

from pulpcore.tests.functional.utils import PulpTaskError
from pulp_rpm.tests.functional.constants import (
    RPM_WITH_NON_ASCII_NAME,
    RPM_WITH_NON_ASCII_URL,
    RPM_WITH_NON_UTF_8_NAME,
    RPM_WITH_NON_UTF_8_URL,
)


"""Test upload of RPMs with different character encoding.

This test targets the following issues:

* `Pulp #4210 <https://pulp.plan.io/issues/4210>`_
* `Pulp #4215 <https://pulp.plan.io/issues/4215>`_
"""


def test_upload_non_ascii(
    tmp_path, artifacts_api_client, http_get, rpm_package_api, monitor_task, delete_orphans_pre
):
    """Test whether one can upload an RPM with non-ascii metadata."""
    temp_file = tmp_path / str(uuid.uuid4())
    temp_file.write_bytes(http_get(RPM_WITH_NON_ASCII_URL))
    artifact = artifacts_api_client.create(temp_file)
    response = rpm_package_api.create(
        artifact=artifact.pulp_href,
        relative_path=RPM_WITH_NON_ASCII_NAME,
    )
    task = monitor_task(response.task)
    assert len(task.created_resources) == 1


def test_upload_non_utf8(
    tmp_path, artifacts_api_client, http_get, rpm_package_api, monitor_task, delete_orphans_pre
):
    """Test whether an exception is raised when non-utf-8 is uploaded."""
    temp_file = tmp_path / str(uuid.uuid4())
    temp_file.write_bytes(http_get(RPM_WITH_NON_UTF_8_URL))
    artifact = artifacts_api_client.create(temp_file)
    with pytest.raises(PulpTaskError) as ctx:
        response = rpm_package_api.create(
            artifact=artifact.pulp_href,
            relative_path=RPM_WITH_NON_UTF_8_NAME,
        )
        monitor_task(response.task)

    assert "'utf-8' codec can't decode byte 0x80 in position 168: invalid start" in str(ctx.value)



# Generated by Django 3.2.13 on 2022-04-21 15:14

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0045_modulemd_fields'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='rpmalternatecontentsource',
            options={'default_related_name': '%(app_label)s_%(model_name)s', 'permissions': [('refresh_rpmalternatecontentsource', 'Refresh an Alternate Content Source'), ('manage_roles_rpmalternatecontentsource', 'Can manage roles on ACS')]},
        ),
        migrations.AlterModelOptions(
            name='rpmdistribution',
            options={'default_related_name': '%(app_label)s_%(model_name)s', 'permissions': [('manage_roles_rpmdistribution', 'Can manage roles on an RPM distribution')]},
        ),
        migrations.AlterModelOptions(
            name='rpmpublication',
            options={'default_related_name': '%(app_label)s_%(model_name)s', 'permissions': [('manage_roles_rpmpublication', 'Can manage roles on an RPM publication')]},
        ),
        migrations.AlterModelOptions(
            name='rpmremote',
            options={'default_related_name': '%(app_label)s_%(model_name)s', 'permissions': [('manage_roles_rpmremote', 'Can manage roles on an RPM remotes')]},
        ),
        migrations.AlterModelOptions(
            name='rpmrepository',
            options={'default_related_name': '%(app_label)s_%(model_name)s', 'permissions': [('manage_roles_rpmrepository', 'Can manage roles on RPM repositories'), ('modify_content_rpmrepository', 'Add content to, or remove content from a repository'), ('repair_rpmrepository', 'Copy a repository'), ('sync_rpmrepository', 'Sync a repository'), ('delete_rpmrepository_version', 'Delete a repository version')]},
        ),
        migrations.AlterModelOptions(
            name='ulnremote',
            options={'default_related_name': '%(app_label)s_%(model_name)s', 'permissions': [('manage_roles_ulnremote', 'Can manage roles on an ULN remotes')]},
        ),
    ]



# Generated by Django 3.2.15 on 2022-08-12 16:28

from django.core.files.storage import default_storage
from django.db import migrations, models, transaction


def convert_artifact_to_snippets(apps, schema_editor):
    """Create snippet from artifact and remove artifact."""
    with transaction.atomic():
        Modulemd = apps.get_model("rpm", "Modulemd")
        ModulemdDefaults = apps.get_model("rpm", "ModulemdDefaults")
        ContentArtifact = apps.get_model("core", "ContentArtifact")
        modules_with_snippet = []
        defaults_with_snippet = []

        for module in Modulemd.objects.all():
            artifact = module._artifacts.get()
            if default_storage.exists(artifact.file.name):
                module.snippet = artifact.file.read().decode("utf-8")
            content_artifact = ContentArtifact.objects.filter(content__pk=module.pk)
            content_artifact.delete()
            modules_with_snippet.append(module)

        for default in ModulemdDefaults.objects.all():
            artifact = default._artifacts.get()
            if default_storage.exists(artifact.file.name):
                default.snippet = artifact.file.read().decode("utf-8")
            content_artifact = ContentArtifact.objects.filter(content__pk=default.pk)
            content_artifact.delete()
            defaults_with_snippet.append(default)

        Modulemd.objects.bulk_update(modules_with_snippet, ["snippet"])
        ModulemdDefaults.objects.bulk_update(defaults_with_snippet, ["snippet"])


class Migration(migrations.Migration):

    dependencies = [
        ("rpm", "0043_textfield_conversion"),
    ]

    operations = [
        migrations.AddField(
            model_name="modulemd",
            name="snippet",
            field=models.TextField(default=""),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="modulemddefaults",
            name="snippet",
            field=models.TextField(default=""),
            preserve_default=False,
        ),
        migrations.RunPython(convert_artifact_to_snippets),
    ]



"""Tests that publish rpm plugin repositories."""
import pytest
from datetime import datetime
from html.parser import HTMLParser
from tempfile import NamedTemporaryFile
from urllib.parse import urljoin

from productmd.treeinfo import TreeInfo

from pulp_rpm.tests.functional.constants import (
    RPM_PACKAGE_CONTENT_NAME,
    CENTOS7_URL,
    CENTOS8_STREAM_APPSTREAM_URL,
    CENTOS8_STREAM_BASEOS_URL,
    EPEL7_URL,
)


class PackagesHtmlParser(HTMLParser):
    """HTML parser that looks for the first link to a file.

    This parser look for the first link that doesn't have a trailing slash. It then sets the
    package_href attribute to the value of the href.
    """

    package_href = None

    def handle_starttag(self, tag, attrs):
        """Looks for a tags that don't end with a slash."""
        if not self.package_href:
            if tag == "a" and not attrs[0][1].endswith("/"):
                self.package_href = attrs[0][1]
        else:
            super().handle_starttag(tag, attrs)


def parse_date_from_string(s, parse_format="%Y-%m-%dT%H:%M:%S.%fZ"):
    """Parse string to datetime object.

    :param s: str like '2018-11-18T21:03:32.493697Z'
    :param parse_format: str defaults to %Y-%m-%dT%H:%M:%S.%fZ
    :return: datetime.datetime
    """
    if isinstance(s, datetime):
        return s
    else:
        return datetime.strptime(s, parse_format)


@pytest.fixture
def centos_8stream_baseos_extra_tests(rpm_distribution_factory, http_get):
    def _extra_test(publication_href):
        # Test that the .treeinfo file is available and AppStream sub-repo is published correctly
        distribution = rpm_distribution_factory(publication=publication_href)
        treeinfo_file = http_get(urljoin(distribution.base_url, ".treeinfo"))
        treeinfo = TreeInfo()
        with NamedTemporaryFile("wb") as temp_file:
            temp_file.write(treeinfo_file)
            temp_file.flush()
            treeinfo.load(f=temp_file.name)
        for variant_name, variant in treeinfo.variants.variants.items():
            if variant_name == "BaseOS":
                assert variant.paths.repository == "."
                assert variant.paths.packages == "Packages"
            elif variant_name == "AppStream":
                assert variant.paths.repository == "AppStream"
                assert variant.paths.packages == "AppStream/Packages"
                # Find the first package in the 'AppStream/Packages/a/' directory and download it
                parser = PackagesHtmlParser()
                a_packages_href = urljoin(
                    distribution.base_url, "{}/a/".format(variant.paths.packages)
                )
                a_packages_listing = http_get(a_packages_href)
                parser.feed(a_packages_listing.__str__())
                full_package_path = urljoin(a_packages_href, parser.package_href)
                http_get(full_package_path)

    return _extra_test


@pytest.mark.parametrize(
    "url,extra_tests",
    [
        (EPEL7_URL, None),
        (CENTOS7_URL, None),
        (CENTOS8_STREAM_BASEOS_URL, "centos_8stream_baseos_extra_tests"),
        (CENTOS8_STREAM_APPSTREAM_URL, None),
        (CENTOS8_STREAM_BASEOS_URL, None),
    ],
)
def test_rpm_publish(
    url,
    extra_tests,
    init_and_sync,
    rpm_repository_version_api,
    rpm_publication_api,
    monitor_task,
    delete_orphans_pre,
    request,
):
    """Publish repositories with the rpm plugin.

    Do the following:

    1. Create a repository and a remote.
    2. Assert that repository version is None.
    3. Sync the remote.
    4. Assert that repository version is not None.
    5. Assert that distribution_tree units were added and are present in the repo.
    6. Publish
    """
    repo, remote, task = init_and_sync(url=url, policy="on_demand", return_task=True)
    created_at = parse_date_from_string(task.pulp_created)
    started_at = parse_date_from_string(task.started_at)
    finished_at = parse_date_from_string(task.finished_at)
    task_duration = finished_at - started_at
    waiting_time = started_at - created_at
    print(
        "\n->     Sync => Waiting time (s): {wait} | Service time (s): {service}".format(
            wait=waiting_time.total_seconds(), service=task_duration.total_seconds()
        )
    )

    # Check that we have the correct content counts.
    repo_ver = rpm_repository_version_api.read(repo.latest_version_href)
    assert RPM_PACKAGE_CONTENT_NAME in repo_ver.content_summary.present.keys()
    assert RPM_PACKAGE_CONTENT_NAME in repo_ver.content_summary.added.keys()

    # Publishing
    response = rpm_publication_api.create({"repository": repo.pulp_href})
    task = monitor_task(response.task)
    created_at = parse_date_from_string(task.pulp_created)
    started_at = parse_date_from_string(task.started_at)
    finished_at = parse_date_from_string(task.finished_at)
    task_duration = finished_at - started_at
    waiting_time = started_at - created_at
    print(
        "\n->     Publish => Waiting time (s): {wait} | Service time (s): {service}".format(
            wait=waiting_time.total_seconds(), service=task_duration.total_seconds()
        )
    )
    if extra_tests:
        # fixtures and parameterization don't like each other - this hack gets us around that
        xtra_fn = request.getfixturevalue(extra_tests)
        xtra_fn(task.created_resources[0])



from pulpcore.plugin.replica import Replicator

from pulp_glue.rpm.context import (
    PulpRpmDistributionContext,
    PulpRpmPublicationContext,
    PulpRpmRepositoryContext,
)


from pulp_rpm.app.models import RpmDistribution, RpmRemote, RpmRepository
from pulp_rpm.app.tasks import synchronize as rpm_synchronize


class RpmReplicator(Replicator):
    repository_ctx_cls = PulpRpmRepositoryContext
    distribution_ctx_cls = PulpRpmDistributionContext
    publication_ctx_cls = PulpRpmPublicationContext
    app_label = "rpm"
    remote_model_cls = RpmRemote
    repository_model_cls = RpmRepository
    distribution_model_cls = RpmDistribution
    distribution_serializer_name = "RpmDistributionSerializer"
    repository_serializer_name = "RpmRepositorySerializer"
    remote_serializer_name = "RpmRemoteSerializer"
    sync_task = rpm_synchronize

    def repository_extra_fields(self, remote):
        """Returns a dictionary where each key is a field on an RpmRemote."""
        return dict(autopublish=False)

    def sync_params(self, repository, remote):
        """Returns a dictionary where key is a parameter for the sync task."""
        return dict(
            remote_pk=remote.pk,
            repository_pk=repository.pk,
            sync_policy="mirror_complete",
            skip_types=[],
            optimize=True,
        )


REPLICATION_ORDER = [RpmReplicator]



"""Tests that verify download of content served by Pulp."""
import pytest

from pulp_rpm.tests.functional.constants import DOWNLOAD_POLICIES
from pulpcore.client.pulp_rpm import RpmRpmPublication


@pytest.mark.parallel
@pytest.mark.parametrize("policy", DOWNLOAD_POLICIES)
def test_pulp_pulp_sync(
    policy,
    init_and_sync,
    rpm_repository_version_api,
    rpm_publication_api,
    rpm_distribution_factory,
    gen_object_with_cleanup,
):
    """Verify whether content served by Pulp can be synced.

    The initial sync to Pulp is one of many different download policies, the second sync is
    immediate in order to exercise downloading all of the files.

    Do the following:

    1. Create, populate, publish, and distribute a repository.
    2. Sync other repository using as remote url,
    the distribution base_url from the previous repository.

    """
    repo, remote = init_and_sync(policy=policy)

    # Create a publication.
    publish_data = RpmRpmPublication(
        repository=repo.pulp_href,
        metadata_checksum_type="sha384",
        package_checksum_type="sha224",
    )
    publication = gen_object_with_cleanup(rpm_publication_api, publish_data)

    # Create a distribution.
    distribution = rpm_distribution_factory(publication=publication.pulp_href)

    # Create another repo pointing to distribution base_url
    # Should this second policy always be "immediate"?
    repo2, remote2 = init_and_sync(url=distribution.base_url, policy="immediate")

    repo_ver = rpm_repository_version_api.read(repo.latest_version_href)
    summary = {k: v["count"] for k, v in repo_ver.content_summary.present.items()}
    repo_ver2 = rpm_repository_version_api.read(repo2.latest_version_href)
    summary2 = {k: v["count"] for k, v in repo_ver2.content_summary.present.items()}
    assert summary == summary2

    added = {k: v["count"] for k, v in repo_ver.content_summary.added.items()}
    added2 = {k: v["count"] for k, v in repo_ver2.content_summary.added.items()}
    assert added == added2



from .advisory import (  # noqa
    UpdateCollection,
    UpdateCollectionPackage,
    UpdateRecord,
    UpdateReference,
)
from .comps import PackageCategory, PackageEnvironment, PackageGroup, PackageLangpacks  # noqa
from .custom_metadata import RepoMetadataFile  # noqa
from .distribution import Addon, Checksum, DistributionTree, Image, Variant  # noqa
from .modulemd import Modulemd, ModulemdDefaults, ModulemdObsolete  # noqa
from .package import Package, format_nevra, format_nevra_short, format_nvra  # noqa
from .repository import RpmDistribution, RpmPublication, RpmRemote, UlnRemote, RpmRepository  # noqa

# at the end to avoid circular import as ACS needs import RpmRemote
from .acs import RpmAlternateContentSource  # noqa
